""" Unit tests for the the 'sound.parabolic_equation' module"""
import pytest
import os
import numpy as np
import kadlu.sound.parabolic_equation as pe 
from kadlu.utils import deg2rad

def test_prop_defr():
    """ Check that defractive propagation matrix is computed correctly """
    dz = 50
    z = np.array([0,50,100,150,200,-150,-100,-50], dtype=float)
    kz = z * 2 * np.pi / (len(z) * dz**2) 
    k0 = 2 * np.pi * 10 / 1500
    UD = pe.prop_defr(x=5, k0=k0, kz=kz, nq=6)
    assert UD.shape == (1,8,6)
    answ = np.array([1.+0.j, 0.9998832 -0.01528329j, 0.99748706-0.07084889j,
                     0.87806263-0.18663797j, 0.77394444-0.16450697j, 0.87806263-0.18663797j,
                     0.99748706-0.07084889j, 0.9998832 -0.01528329j])
    assert np.all(np.abs(UD[0,:,0] - answ) < 1e-7)

def test_thomson_starter():
    """ Check that thomson starter field has expected shape and values"""
    grid = pe.Grid(100., 1000., 100., 500., 10.*np.pi/180.)
    psi = pe.thomson_starter(k0=0.04, kz=grid.kz, dz=grid.dz, zs=9, theta1=86*np.pi/180.)
    assert psi.shape == (1,grid.nz,1)
    psi = np.round(psi[0], 4)
    answ = np.array([[ 0.    +0.j    ],
            [ 0.0101-0.0101j],
            [ 0.0205-0.0205j],
            [ 0.0319-0.0319j],
            [ 0.0451-0.0451j],
            [ 0.    +0.j    ],
            [-0.0451+0.0451j],
            [-0.0319+0.0319j],
            [-0.0205+0.0205j],
            [-0.0101+0.0101j]])
    assert np.all(answ == psi)

def test_thomson_starter_multiple_depths():
    """ Check that thomson starter field has expected shape and values 
        for multiple depth values """
    grid = pe.Grid(100., 1000., 100., 500., 10.*np.pi/180.)
    psi = pe.thomson_starter(k0=0.04, kz=grid.kz, dz=grid.dz, zs=[9, 99], theta1=86*np.pi/180.)
    psi = np.round(psi, 4)
    answ0 = np.array([[0.+0.j],[0.0101-0.0101j],[0.0205-0.0205j],\
        [0.0319-0.0319j],[ 0.0451-0.0451j],[0.+0.j],[-0.0451+0.0451j],\
        [-0.0319+0.0319j],[-0.0205+0.0205j],[-0.0101+0.0101j]])
    assert np.all(psi[0,:,:] == answ0)
    answ1 = np.array([[0.+0.j],[0.1039-0.1039j],[0.1723-0.1723j],\
        [0.1806-0.1806j],[0.1222-0.1222j],[0.+0.j],[-0.1222+0.1222j],\
        [-0.1806+0.1806j],[-0.1723+0.1723j],[-0.1039+0.1039j]])
    assert np.all(psi[1,:,:] == answ1)

def test_create_grid():
    """ Check that the created grid is as expected"""
    g = pe.Grid(dr=100, r_max=1000, dq=45*np.pi/180, dz=100, z_max=500)
    # radial
    assert g.dr == 100
    assert g.nr == 11
    answ = np.array([0, 100, 200, 300, 400, 500, 600, 700, 800, 900, 1000])
    assert np.all(g.r == answ)
    # azimuthal
    assert g.dq == 45*np.pi/180
    assert g.nq == 8
    answ = np.array([0, 45, 90, 135, -180, -135, -90, -45]) * deg2rad
    assert np.all(np.abs(g.q - answ) < 1E-9)
    # vertical
    assert g.dz == 100
    assert g.nz == 10
    answ = np.array([0., 100., 200., 300., 400., 500., -400., -300., -200., -100.])
    assert np.all(g.z == answ)
    # below/above indices
    answ = np.array([0, 1, 2, 3, 4, 5])
    assert np.all(g.below == answ)
    answ = np.array([6, 7, 8, 9])
    assert np.all(g._above == answ)
    answ = np.array([4, 3, 2, 1])
    assert np.all(g._mirror == answ)

def test_create_grid_selected_angles():
    """ Check that we can create a grid for selected angular slices"""
    g = pe.Grid(dr=100, r_max=1000, q=[45*deg2rad, 120*deg2rad], dz=100, z_max=500)
    assert g.dq == None
    assert g.nq == 2
    answ = np.array([45*deg2rad, 120*deg2rad])
    assert np.all(np.abs(g.q - answ) < 1E-9)

def test_mirror_grid():
    """ Check that the mirror method returns the expected array"""
    g = pe.Grid(dr=100, r_max=1000, dq=45*deg2rad, dz=100, z_max=500)
    # mirror 1d array
    a = np.array([6.0, 5.0, 4.0, 3.0, 2.0, 1.0, np.nan, np.nan, np.nan, np.nan])
    answ = np.array([6.0, 5.0, 4.0, 3.0, 2.0, 1.0, 2.0, 3.0, 4.0, 5.0])
    a = g.mirror(a)
    assert np.all(a == answ)
    # mirror 2d array
    a = np.array([[6.0], [5.0], [4.0], [3.0], [2.0], [1.0], [np.nan], [np.nan], [np.nan], [np.nan]])
    a = np.swapaxes(a, 0, 1)
    answ = np.array([[6.0], [5.0], [4.0], [3.0], [2.0], [1.0], [2.0], [3.0], [4.0], [5.0]])
    answ = np.swapaxes(answ, 0, 1)
    a = g.mirror(a, z_axis=1)
    assert np.all(a == answ)

def test_eff_index_refr_sq_flat_seafloor():
    """ Check that the effective index of refraction is computed 
        correctly in an environment with uniform seafloor depth """
    z    = np.array([[0,0],[50,50],[100,100],[150,150],[-100,-100],[-50,-50]])
    zb   = np.ones(z.shape) * 90
    dzb2 = np.zeros(z.shape)
    k0   = 0.041887902047863905
    L    = 75
    r    = 1.0
    rb   = 1.5
    n2   = np.ones(z.shape) * 1.0272096420745072
    n2b  = 0.7783506614961636+0.014264417521561993j
    n2_eff = pe.eff_index_refr_sq(z,zb,dzb2,k0,L,n2,n2b,r,rb)
    assert n2_eff.shape == z.shape
    answ = np.array([[1.03298621+0.j, 1.03298621+0.j],
                     [1.03339092+0.j, 1.03339092+0.j],
                     [0.77299581+0.01426442j, 0.77299581+0.01426442j],
                     [0.77096953+0.01426442j, 0.77096953+0.01426442j],
                     [0.77299581+0.01426442j, 0.77299581+0.01426442j],
                     [1.03339092+0.j, 1.03339092+0.j]])
    assert np.all(np.abs(n2_eff - answ) < 1e-7)

def test_eff_index_refr_sq_sloping_seafloor():
    """ Check that the effective index of refraction is computed 
        correctly in an environment with a sloping seafloor """
    z    = np.array([[0,0],[50,50],[100,100],[150,150],[-100,-100],[-50,-50]])
    zb   = np.ones(z.shape) * 90
    dzb  = np.ones(z.shape) * 0.45
    dzb2 = dzb**2
    k0   = 0.041887902047863905
    L    = 75
    r    = 1.0
    rb   = 1.5
    n2   = np.ones(z.shape) * 1.0272096420745072
    n2b  = 0.7783506614961636+0.014264417521561993j
    n2_eff = pe.eff_index_refr_sq(z,zb,dzb2,k0,L,n2,n2b,r,rb)
    assert n2_eff.shape == z.shape
    answ = np.array([[1.03415596+0.j, 1.03415596+0.j],
                    [1.03464263+0.j, 1.03464263+0.j],
                    [0.77191145+0.01426442j, 0.77191145+0.01426442j],
                    [0.76947486+0.01426442j, 0.76947486+0.01426442j],
                    [0.77191145+0.01426442j, 0.77191145+0.01426442j],
                    [1.03464263+0.j, 1.03464263+0.j]])
    assert np.all(np.abs(n2_eff - answ) < 1e-7)

def test_index_refr_sq():
    """ Check that the index of refraction squared is computed correctly 
        including complex attenuation term"""
    n2 = pe.index_refr_sq(c0=1500, c=1700, alpha=0.5)
    assert np.abs(n2 - (0.7783506614961636+0.01426441752156199j)) < 1e-9

def test_init_transm_loss():
    """ Check that we can initialize and instance of the PESolver
        with a flat seafloor and uniform sound speed """
    freq = 100
    prop_range = 40
    angular_bin = 12
    bottom = {'sound_speed':1700,'density':1.5,'attenuation':0.5}
    def bathy_func(x,y,grid=None): return 430*np.ones(x.shape)
    def bathy_deriv_func(x,y,axis): return np.zeros(x.shape)
    def sound_speed_func(x,y,z): return 1480*np.ones(x.shape)   
    transm_loss = pe.TransmissionLoss(freq=freq, bathy_func=bathy_func, 
        bathy_deriv_func=bathy_deriv_func, sound_speed_func=sound_speed_func, 
        bottom=bottom, propagation_range=prop_range, angular_bin=angular_bin)
    assert transm_loss.c0 == 1500
    assert transm_loss.k0 == 2 * np.pi / transm_loss.c0 * freq
    assert transm_loss.grid.dr == transm_loss.c0 / freq
    assert transm_loss.grid.dz == transm_loss.c0 / freq / 2.
    assert transm_loss.grid.dq == pytest.approx(angular_bin * np.pi/180, abs=1e-6)
    assert np.abs(np.max(transm_loss.grid.r) - 1e3*prop_range) < transm_loss.grid.dr
    z_max = 4/3 * (430 + 3 * transm_loss.c0 / freq)
    assert np.abs(np.max(transm_loss.grid.z) - z_max) < transm_loss.grid.dz
    # now, try with kwargs
    transm_loss = pe.TransmissionLoss(freq=freq, bathy_func=bathy_func, 
        bathy_deriv_func=bathy_deriv_func, sound_speed_func=sound_speed_func, 
        bottom=bottom, propagation_range=prop_range, angular_bin=angular_bin,
        dq=24*np.pi/180)
    assert transm_loss.grid.dq == pytest.approx(24 * np.pi/180, abs=1e-6)

def test_transm_loss_solve_pe():
    freq = 100
    prop_range = 0.2
    angular_bin = 180
    grid_kwargs = {'dr':50, 'dz':50, 'z_max':150}
    bottom = {'sound_speed':1700,'density':1.5,'attenuation':0.5}
    source_depth = 16
    def bathy_func(x,y,grid=None): return 90*np.ones(x.shape)
    def bathy_deriv_func(x,y,axis): return np.zeros(x.shape)
    def sound_speed_func(x,y,z): return 1480*np.ones(x.shape)   
    transm_loss = pe.TransmissionLoss(freq=freq, bathy_func=bathy_func, 
        bathy_deriv_func=bathy_deriv_func, sound_speed_func=sound_speed_func, 
        bottom=bottom, propagation_range=prop_range, angular_bin=angular_bin,
        **grid_kwargs)
    transm_loss._init_output(num_sources=1, rec_depth=np.array([.1]))        
    psi = transm_loss._solve_pe(source_depth=source_depth, rec_depth=[.1], return_field=True, progress_bar=False, aperture=86)
    answ = np.array([[[0+0j, 0+0j],
             [7.25511049e-02-5.22233517e-03j, 7.25511049e-02-5.22233517e-03j],
             [7.41434583e-02+1.77258038e-02j, 7.41434583e-02+1.77258038e-02j],
             [0+0j, 0+0j],
             [-7.41434583e-02-1.77258038e-02j, -7.41434583e-02-1.77258038e-02j],
             [-7.25511049e-02+5.22233517e-03j, -7.25511049e-02+5.22233517e-03j]]])
    assert np.all(np.abs(psi - answ) < 1e-9*np.max(np.abs(answ)))

def test_compute_transm_loss_flat_seafloor_uniform_ssp():
    """ Check that PESolver returns expected transmission loss for 
        flat bathymetry and uniform sound speed profile"""
    freq = 10
    prop_range = 10
    angular_bin = 45
    grid_kwargs = {'dr':1000, 'dz':1000}
    bottom = {'sound_speed':1700,'density':1.5,'attenuation':0.5}
    source_depth = 9900
    def bathy_func(x,y,grid=None): return 10000*np.ones(x.shape)
    def bathy_deriv_func(x,y,axis): return np.zeros(x.shape)
    def sound_speed_func(x,y,z): return 1500*np.ones(x.shape)   
    transm_loss = pe.TransmissionLoss(freq=freq, bathy_func=bathy_func, 
        bathy_deriv_func=bathy_deriv_func, sound_speed_func=sound_speed_func, 
        bottom=bottom, propagation_range=prop_range, angular_bin=angular_bin,
        **grid_kwargs)
    tl, axes = transm_loss.calc(source_depth=source_depth, rec_depth=[.1], aperture=88, progress_bar=False)
    expected = np.array([[-164.6146, -170.7267, -176.8872, -172.0097, -182.3044, -176.6344, -176.9115, -183.7996, -177.9538, -181.3532],
                         [-164.6146, -170.7267, -176.8872, -172.0097, -182.3044, -176.6344, -176.9115, -183.7996, -177.9538, -181.3532],
                         [-164.6146, -170.7267, -176.8872, -172.0097, -182.3044, -176.6344, -176.9115, -183.7996, -177.9538, -181.3532],
                         [-164.6146, -170.7267, -176.8872, -172.0097, -182.3044, -176.6344, -176.9115, -183.7996, -177.9538, -181.3532],
                         [-164.6146, -170.7267, -176.8872, -172.0097, -182.3044, -176.6344, -176.9115, -183.7996, -177.9538, -181.3532],
                         [-164.6146, -170.7267, -176.8872, -172.0097, -182.3044, -176.6344, -176.9115, -183.7996, -177.9538, -181.3532],
                         [-164.6146, -170.7267, -176.8872, -172.0097, -182.3044, -176.6344, -176.9115, -183.7996, -177.9538, -181.3532],
                         [-164.6146, -170.7267, -176.8872, -172.0097, -182.3044, -176.6344, -176.9115, -183.7996, -177.9538, -181.3532]])
    np.testing.assert_array_almost_equal(-tl[0,0], expected, decimal=3) 

def test_compute_transm_loss_multiple_sources():
    """ Check that PESolver can handle multiple source depths"""
    freq = 10
    prop_range = 10
    angular_bin = 45
    grid_kwargs = {'dr':1000, 'dz':1000}
    bottom = {'sound_speed':1700,'density':1.5,'attenuation':0.5}
    def bathy_func(x,y,grid=None): return 10000*np.ones(x.shape)
    def bathy_deriv_func(x,y,axis): return np.zeros(x.shape)
    def sound_speed_func(x,y,z): return 1500*np.ones(x.shape)   
    transm_loss = pe.TransmissionLoss(freq=freq, bathy_func=bathy_func, 
        bathy_deriv_func=bathy_deriv_func, sound_speed_func=sound_speed_func, 
        bottom=bottom, propagation_range=prop_range, angular_bin=angular_bin,
        **grid_kwargs)
    tl0, ax  = transm_loss.calc(source_depth=9900, rec_depth=[.1], aperture=88, progress_bar=False)
    tl1, ax  = transm_loss.calc(source_depth=8800, rec_depth=[.1], aperture=88, progress_bar=False)
    tl01, ax = transm_loss.calc(source_depth=[9900,8800], rec_depth=[.1], aperture=88, progress_bar=False)
    np.testing.assert_array_almost_equal(tl01[0,0], tl0[0,0], decimal=3) 
    np.testing.assert_array_almost_equal(tl01[1,0], tl1[0,0], decimal=3) 



""" Unit tests for the the 'sound.geophony' module in the 'kadlu' package

    Authors: Oliver Kirsebom
    contact: oliver.kirsebom@dal.ca
    Organization: MERIDIAN-Intitute for Big Data Analytics
    Team: Acoustic data Analytics, Dalhousie University
    Project: packages/kadlu
             Project goal: Tools for underwater soundscape modeling
     
    License:

"""

import pytest
import os
import numpy as np
from kadlu.sound.geophony import geophony, transmission_loss, kewley_sl_func, source_level
from kadlu.geospatial.ocean import Ocean
from kadlu.utils import R1_IUGG, deg2rad

current_dir = os.path.dirname(os.path.realpath(__file__))
path_to_assets = os.path.join(os.path.dirname(current_dir),"assets")

def test_kewley_sl_func():
    sl1 = kewley_sl_func(freq=10, wind_uv=0)
    sl2 = kewley_sl_func(freq=40, wind_uv=2.57)
    assert sl1 == sl2
    assert sl2 == 40.0
    sl3 = kewley_sl_func(freq=40, wind_uv=5.14)
    assert sl3 == 44.0
    sl4 = kewley_sl_func(freq=100, wind_uv=5.14)
    assert sl4 == 42.5

def test_source_level():
    ok = {'load_bathymetry': 10000, 'load_wind_uv': 5.14}
    o = Ocean(**ok)
    sl = source_level(freq=10, x=0, y=0, area=1, ocean=o, sl_func=kewley_sl_func)
    assert sl == 44.0
    sl = source_level(freq=100, x=[0,100], y=[0,100], area=[1,2], ocean=o, sl_func=kewley_sl_func)
    assert sl[0] == 42.5
    assert sl[1] == sl[0] + 10*np.log10(2)

def test_geophony_flat_seafloor():
    """ Check that we can execute the geophony method for a 
        flat seafloor and uniform sound speed profile"""
    kwargs = {'load_bathymetry':10000, 'load_wind_uv':1.0, 'ssp':1480, 'angular_bin':90, 'dr':1000, 'dz':1000}
    geo = geophony(freq=100, south=44, north=46, west=-60, east=-58, depth=[100, 2000], xy_res=71, **kwargs)
    spl = geo['spl']
    x = geo['x']
    y = geo['y']
    assert x.shape[0] == 3
    assert y.shape[0] == 5
    assert spl.shape[0] == 3
    assert spl.shape[1] == 5
    assert spl.shape[2] == 2
    assert np.all(np.diff(x) == 71e3)
    assert np.all(np.diff(y) == 71e3)
    # try again, but this time for specific location
    kwargs = {'load_bathymetry':10000, 'load_wind_uv':1.0, 'ssp':1480, 'angular_bin':90, 'dr':1000, 'dz':1000, 'propagation_range':50}
    geo = geophony(freq=100, lat=45, lon=-59, depth=[100, 2000], **kwargs)

def test_geophony_in_canyon(bathy_canyon):
    """ Check that we can execute the geophony method for a 
        canyon-shaped bathymetry and uniform sound speed profile"""
    kwargs = {'load_bathymetry':bathy_canyon, 'load_wind_uv':1.0, 'ssp':1480, 'angular_bin':90, 'dr':1000, 'dz':1000}
    z = [100, 1500, 3000]
    geo = geophony(freq=10, south=43, north=46, west=60, east=62, depth=z, xy_res=71, **kwargs)
    spl = geo['spl']
    x = geo['x']
    y = geo['y']
    assert spl.shape[0] == x.shape[0]
    assert spl.shape[1] == y.shape[0]
    assert spl.shape[2] == len(z)
    assert np.all(np.diff(x) == 71e3)
    assert np.all(np.diff(y) == 71e3)    
    # check that noise is NaN below seafloor and non Nan above
    bathy = np.swapaxes(np.reshape(geo['bathy'], newshape=(y.shape[0], x.shape[0])), 0, 1)
    bathy = bathy[:,:,np.newaxis]
    xyz = np.ones(shape=bathy.shape) * z
    idx = np.nonzero(xyz >= bathy)
    assert np.all(np.isnan(spl[idx]))
    idx = np.nonzero(xyz < bathy)
    assert np.all(~np.isnan(spl[idx]))

def test_transmission_loss_real_world_env():
    """ Check that we can initialize a transmission loss object 
        for a real-world environment and obtain the expected result """

    from datetime import datetime
    bounds = dict(
               start=datetime(2015,1,1), end=datetime(2015,1,2), 
               top=0, bottom=10000
             )    
    src = dict(load_bathymetry='gebco', load_temp='hycom', load_salinity='hycom')
    sound_source = {'freq': 200, 'lat': 43.8, 'lon': -59.04, 'source_depth': 12}
    seafloor = {'sound_speed':1700,'density':1.5,'attenuation':0.5}
    transm_loss, ocean = transmission_loss(seafloor=seafloor, propagation_range=20, **src, **bounds, **sound_source, dr=100, angular_bin=45, dz=50, return_ocean=True)

    # check fetching and interpolation of ocean variables
    lats = np.linspace(43.7, 43.9, num=3)
    lons = np.linspace(-59.14, -58.94, num=3)
    seafloor_depth = ocean.bathy(lat=lats, lon=lons, grid=True)
    max_depth = 1712.83265634
    depths = np.linspace(0, max_depth, num=3)
    temp = ocean.temp(lat=lats, lon=lons, depth=depths, grid=True)
    salinity = ocean.salinity(lat=lats, lon=lons, depth=depths, grid=True)  

#    answ_seafloor_depth = np.array([[ 264.37,  579.97, 1396.33],
#                                    [  99.07,  283.93, 1712.83],
#                                    [  79.96,  430.3,  1155.  ]])
    answ_seafloor_depth = np.array([[ 267.75,  606.44, 1398.15],
                                    [  99.94,  273.12, 1651.99],
                                    [  80.17,  370.65, 1307.41]])

    answ_temp = np.array([[[ 7.32,  4.67,  4.67],
                           [ 5.49, 10.,   10.  ],
                           [ 3.98,  9.36,  9.36]],
                          [[ 7.35,  4.04,  3.99],
                           [ 6.3,   4.47,  4.47],
                           [ 3.88,  5.36,  5.36]],
                          [[ 5.66,  4.17,  3.72],
                           [ 6.83,  4.07,  3.76],
                           [ 4.5,   4.31,  4.31]]])

    answ_salinity = np.array([[[33.3,  34.97, 34.97],
                               [32.52, 34.79, 34.79],
                               [31.71, 34.32, 34.32]],
                              [[33.31, 34.95, 34.95],
                               [33.,   34.95, 34.95],
                               [31.77, 34.94, 34.94]],
                              [[32.96, 34.95, 34.96],
                               [33.23, 34.95, 34.95],
                               [32.1,  34.95, 34.95]]])

    ##print(temp.round(2))

    np.testing.assert_array_almost_equal(seafloor_depth, answ_seafloor_depth, decimal=2) 

    np.testing.assert_array_almost_equal(temp, answ_temp, decimal=2) 
    np.testing.assert_array_almost_equal(salinity, answ_salinity, decimal=2) 

    # check transmission loss calculation results
    tl_h, ax_h, tl_v, ax_v = transm_loss.calc(vertical=True)

    answ_h = np.array([[97.6, 117.2, 114.9, 120.4, 119.0, 120.1, 121.1, 121.9, 122.7, 123.3], 
                        [97.6, 116.4, 117.7, 117.6, 120.2, 122.0, 121.4, 123.7, 120.1, 122.7], 
                        [97.6, 114.3, 117.2, 118.4, 120.6, 120.2, 122.2, 120.4, 125.3, 121.1], 
                        [97.6, 117.6, 117.9, 117.8, 117.7, 121.1, 119.9, 120.7, 122.4, 123.6], 
                        [97.6, 114.7, 117.5, 117.5, 116.9, 120.1, 121.7, 120.1, 119.9, 121.3], 
                        [97.6, 110.4, 115.5, 119.6, 119.0, 117.9, 117.7, 123.4, 120.7, 122.3], 
                        [97.6, 110.2, 114.9, 117.7, 123.5, 120.1, 121.0, 121.5, 123.5, 120.4], 
                        [97.6, 112.0, 115.6, 118.1, 119.1, 120.2, 121.2, 122.0, 122.8, 123.5]])

    answ_v = np.array([[31.9, 65.4, 68.1, 70.1, 71.1, 72.2, 73.2, 74.0, 74.8, 75.5, 76.1], 
                        [53.4, 147.1, 144.3, 151.1, 145.6, 150.4, 150.7, 150.0, 150.5, 151.3, 152.0], 
                        [59.5, 164.6, 161.3, 168.0, 162.7, 167.4, 167.7, 167.0, 167.5, 168.3, 169.0], 
                        [63.7, 174.9, 171.6, 178.2, 173.0, 177.7, 178.0, 177.3, 177.7, 178.6, 179.2], 
                        [67.6, 182.5, 179.2, 185.9, 180.6, 185.3, 185.6, 184.9, 185.4, 186.2, 186.9], 
                        [71.8, 189.1, 185.9, 192.5, 187.3, 192.0, 192.3, 191.6, 192.1, 192.9, 193.6], 
                        [77.9, 194.6, 193.3, 201.1, 196.7, 200.7, 201.4, 201.7, 202.3, 203.1, 203.7], 
                        [98.9, 213.8, 214.7, 222.8, 219.7, 222.9, 223.8, 224.6, 225.3, 226.0, 226.7]])

    np.testing.assert_array_almost_equal(tl_h[0,0,:,::20], answ_h, decimal=1) 
    np.testing.assert_array_almost_equal(tl_v[0,1::10,::20,0], answ_v, decimal=1) 
    assert tl_h.shape == (1,1,8,200), f'tl_h.shape = {tl_h.shape}'
    assert tl_v.shape == (1,72,201,8), f'tl_v.shape = {tl_v.shape}'


def test_transmission_loss_flat_seafloor():
    """ Check that we can initialize a transmission loss object 
        for a flat seafloor and uniform sound speed profile """
    transm_loss = transmission_loss(freq=100, source_depth=75, propagation_range=0.5, load_bathymetry=2000, ssp=1480, angular_bin=10)
    tl_h, ax_h, tl_v, ax_v = transm_loss.calc(vertical=True)
    answ = np.genfromtxt(os.path.join(path_to_assets, 'lloyd_mirror_f100Hz_SD75m.csv'), delimiter=",")
    assert answ.shape == tl_v[0,:,:,0].shape
    np.testing.assert_array_almost_equal(-tl_v[0,1:,:,0], answ[1:,:], decimal=3) 

#def test_test():
#    from datetime import datetime
#    bounds = dict(
#               south=43.53, north=44.29, west=-59.84, east=-58.48,
#               start=datetime(2015,1,1), end=datetime(2015,1,2), 
#               top=0, bottom=10000
#             )    
#    src = dict(load_bathymetry='chs', load_temp='hycom', load_salinity='hycom')
#    sound_source = {'freq': 200, 'lat': 43.8, 'lon': -59.04, 'source_depth': 12}
#    o = Ocean(**src, **bounds)
#    seafloor = {'sound_speed':1700,'density':1.5,'attenuation':0.5}
#    transm_loss = transmission_loss(seafloor=seafloor, propagation_range=20, **src, **bounds, **sound_source, ssp=1480, dz=50)
#    transm_loss.calc(vertical=False)



import pytest
import kadlu
from datetime import datetime, timedelta

# gulf st lawrence
kwargs = dict(
        south=46.1, west=-64.4, north=47, east=-63.4,
        start=datetime(2016, 2, 1), end=datetime(2016, 2, 1),
        top=0, bottom=5000,
    )

def test_era5_load_windwaveswellheight():
    val, lat, lon, time = kadlu.load(source='era5', var='waveheight', **kwargs)
    assert(len(val)==len(lat)==len(lon))
    assert(len(lat) > 0)

def test_era5_load_wavedirection():
    val, lat, lon, time = kadlu.load(source='era5', var='wavedir', **kwargs)
    assert(len(val)==len(lat)==len(lon))
    assert(len(lat) > 0)

def test_era5_load_waveperiod():
    val, lat, lon, time = kadlu.load(source='era5', var='waveperiod', **kwargs)
    assert(len(val)==len(lat)==len(lon))
    assert(len(lat) > 0)

def test_era5_load_wind():
    ns_offset = 1
    ew_offset = 1

    uval,lat,lon,epoch = kadlu.load(source='era5', var='wind_u', 
                        start=datetime(2016, 3, 9) , end=datetime(2016,3,11),
                        south=44.5541333 - ns_offset, west=-64.17682 - ew_offset, 
                        north=44.5541333 + ns_offset, east=-64.17682 + ew_offset, 
                        top=0, bottom=0)
    assert(len(uval)==len(lat)==len(lon))
    assert len(uval) > 0

    vval,lat,lon,epoch = kadlu.load(source='era5', var='wind_v', 
                        start=datetime(2016, 3, 9) , end=datetime(2016,3,11),
                        south=44.5541333 - ns_offset, west=-64.17682 - ew_offset, 
                        north=44.5541333 + ns_offset, east=-64.17682 + ew_offset, 
                        top=0, bottom=0)
    assert(len(vval)==len(lat)==len(lon))
    assert len(vval) > 0

    uvval,lat,lon,epoch = kadlu.load(source='era5', var='wind_uv', 
                        start=datetime(2016, 3, 9) , end=datetime(2016,3,11),
                        south=44.5541333 - ns_offset, west=-64.17682 - ew_offset, 
                        north=44.5541333 + ns_offset, east=-64.17682 + ew_offset, 
                        top=0, bottom=0)

    assert(len(uvval)==len(lat)==len(lon))
    assert len(uvval) > 0




""" The ocean module provides an interface to fetching, loading 
    and interpolating ocean variables.
"""
import os
import logging
from datetime import timedelta
#from multiprocessing import Process, Queue
from queue import Queue

import numpy as np
from scipy.interpolate import NearestNDInterpolator

import kadlu
from kadlu import index
from kadlu.geospatial.interpolation             import (
        Interpolator2D,
        Interpolator3D,
        Uniform2D,
        Uniform3D,
    )
from kadlu.geospatial.data_sources.data_util    import (
        dt_2_epoch,
        fmt_coords,
        reshape_2D,
        reshape_3D,
        storage_cfg,
        verbosity,
    )
from kadlu.geospatial.data_sources.source_map   import (
        default_val,
        load_map,
        precip_type_map,
        var3d,
    )
from kadlu.geospatial.data_sources.hycom        import Hycom
from kadlu.geospatial.data_sources.era5         import Era5
from kadlu.geospatial.data_sources.wwiii        import Wwiii
from kadlu.utils import center_point


def worker(interpfcn, reshapefcn, cols, var, q):
    """ compute interpolation in parallel worker process

        interpfcn:
            callback function for interpolation
        reshapefcn:
            callback function for reshaping row data into matrix format
            for interpolation
        cols:
            data as returned from load function
        var:
            variable type. used as key in Ocean().interps dictionary
        q:
            shared queue object to pass interpolation back to parent
    """
    if verbosity() and isinstance(cols[0], (int, float)): logging.info(f'OCEAN {var.upper()}{"".join([" " for _ in range(11-len(var))])} loaded uniform value of {cols[0]} for interpolation')
    elif verbosity(): logging.info(f'OCEAN {var.upper()}{ "".join([" " for _ in range(11-len(var)) ]) } loaded {len(cols[0])} points for interpolation')
    obj = interpfcn(**reshapefcn(cols))
    q.put((var, obj))
    return


class Ocean():
    """ class for handling ocean data requests 

        data will be loaded using the given data sources and boundaries
        from arguments. an interpolation for each variable will be computed in 
        parallel

        data will be averaged over time frames for interpolation. for finer
        temporal resolution, define smaller time boundaries

        any of the below load_args may also accept a callback function instead
        of a string or array value if you wish to write your own data loading
        function. the boundary arguments supplied here will be passed to the 
        callable, i.e. north, south, west, east, top, bottom, start, end

        callables or array arguments must be ordered by [val, lat, lon] for 2D 
        data, or [val, lat, lon, depth] for 3D data

        args:
            north, south:
                latitude boundaries (float)
            west, east:
                longitude boundaries (float)
            top, bottom:
                depth range in metres (float)
                only applies to salinity and temperature
            start, end:
                time range for data load query (datetime)
                if multiple times exist within range, they will be averaged
                before computing interpolation
            **loadvars:
                keyword args supplied as 'load_{v}' where v is either a integer,
                float, array of [val, lat, lon[, time[, depth]]], or a data source
                string as described by the source_map

        attrs:
            interps: dict
                Dictionary of data interpolators
            origin: tuple(float, float)
                Latitude and longitude coordinates of the centre point of the 
                geographic bounding box. This point serves as the origin of the 
                planar x-y coordinate system.
            boundaries: dict
                Bounding box for the ocean volume in space and time
    """

    def __init__(self, *, 
            south=kadlu.defaults['south'],   west=kadlu.defaults['west'], 
            north=kadlu.defaults['north'],   east=kadlu.defaults['east'], 
            bottom=kadlu.defaults['bottom'], top=kadlu.defaults['top'],      
            start=kadlu.defaults['start'],   end=kadlu.defaults['end'], 
            **loadvars,
            ):

        logging.debug(f'''south = {south}, north = {north}, west = {west}, east = {east}, top = {top}, bottom = {bottom}''')

        # legacy support
        if 'load_bathy' in loadvars.keys(): loadvars['load_bathymetry'] = loadvars['load_bathy']; del loadvars['load_bathy']
        if 'load_temp' in loadvars.keys(): loadvars['load_temperature'] = loadvars['load_temp']; del loadvars['load_temp']
        if 'load_bathymetry' not in loadvars.keys(): loadvars['load_bathymetry'] = 'gebco'

        callbacks = []
        vartypes = np.unique([f.rsplit('_', 1)[0] for f in load_map.keys()])
        for key in loadvars.keys(): 
            if not key.lstrip('load_') in vartypes and not key == 'load_precip_type': 
                raise TypeError(f'{key} is not a valid argument. valid datasource args include:\n{", ".join([f"load_{v}" for v in vartypes])}')
        load_args = [loadvars[f'load_{v}'] if f'load_{v}' in loadvars.keys() else 0 for v in vartypes]
        
        # if load_args are not callable, convert it to a callable function
        for v, load_arg, ix in zip(vartypes, load_args, range(len(vartypes))):
            if callable(load_arg): callbacks.append(load_arg)

            elif isinstance(load_arg, str):
                key = f'{v}_{load_arg.lower()}'
                assert key in load_map.keys(), f'no map for {key} in load map: \n{load_map}'
                callbacks.append(load_map[key])
                with index(storagedir=storage_cfg(), south=south, north=north, west=west, east=east, top=top, bottom=bottom, start=start, end=end) as fetchmap:
                    fetchmap(callback=load_map[f'{v}_{load_arg.lower()}'])

            elif isinstance(load_arg, (int, float)):
                callbacks.append(lambda val, south, west, start, top, **_: [val, south, west, dt_2_epoch(start), top])

            elif isinstance(load_arg, (list, tuple, np.ndarray)):
                if len(load_arg) not in (3, 4):
                    raise ValueError(f'invalid array shape for load_{v}. '
                    'arrays must be ordered by [val, lat, lon] for 2D data, or '
                    '[val, lat, lon, depth] for 3D data')
                callbacks.append(lambda val, **_: val)

            else: raise TypeError(f'invalid type for load_{v}. '
                  'valid types include string, float, array, and callable')

        #q = Queue()
        attributes = []


        # prepare data pipeline
        is_3D = [v in var3d for v in vartypes]
        is_arr = [not isinstance(arg, (int, float)) for arg in load_args]
        columns = [fcn(val=val, south=south, north=north, west=west, east=east, top=top, bottom=bottom, start=start, end=end) for fcn,val in zip(callbacks, load_args)]
        intrpmap = [(Uniform2D, Uniform3D), (Interpolator2D, Interpolator3D)]
        reshapers = [reshape_3D if v else reshape_2D for v in is_3D]
        # map interpolations to dictionary
        self.interps = {}
        interpolators = map(lambda x, y: intrpmap[x][y], is_arr, is_3D)
        #interpolations = map(
        #    lambda i,r,c,v,q=q: Process(target=worker, args=(i,r,c,v,q)),
        #    interpolators, reshapers, columns, vartypes
        #)

        # assert that no empty arrays were returned by load function
        for col, var in zip(columns, vartypes):
            if isinstance(col, dict) or isinstance(col[0], (int, float)): continue
            assert len(col[0]) > 0, (
                    f'no data found for {var} in region {fmt_coords(dict(south=south, north=north, west=west, east=east, top=top, bottom=bottom, start=start, end=end))}. '
                    f'consider expanding the region')


        #if not os.environ.get('LOGLEVEL') == 'DEBUG':
        #    for i in interpolations: i.start()
        #    set_attributes() 
        #    for i in interpolations: i.join()

        # debug mode: disable parallelization for nicer stack trace
        #elif os.environ.get('LOGLEVEL') == 'DEBUG':
        #logging.debug('OCEAN DEBUG MSG: parallelization disabled')
        for i,r,c,v in zip(interpolators, reshapers, columns, vartypes):
            logging.debug(f'interpolating {v}')
            logging.debug(f'i = {i}\n r = {r}\nc = {c}\nv = {v}')
            obj = i(**r(c))
            #q.put((v, obj))
            attributes.append((v, obj))
        #set_attributes() 
        #def set_attributes(q):

        self.boundaries = dict(south=south, north=north, west=west, east=east, top=top, bottom=bottom, start=start, end=end)  
        self.origin = center_point(lat=[south, north], lon=[west, east])
        #for v in vartypes: self.interps[v].origin = self.origin
        self.precip_src = loadvars['load_precip_type'] if 'load_precip_type' in loadvars.keys() else None
        #x = 0
        #while x <= len(vartypes):
        while len(attributes) > 0:
            #x = x + 1
            #obj = q.get()
            obj = attributes.pop()
            self.interps[obj[0]] = obj[1]
            setattr(self, obj[0],         eval(f'lambda lat, lon, {"depth," if (obj[0] in var3d) else ""} grid=False, **kw: obj[1].interp(lat, lon, {"depth," if (obj[0] in var3d) else ""} grid, **kw)', dict(obj=obj)))
            setattr(self, f'{obj[0]}_xy', eval(f'lambda x, y,     {"z,"     if (obj[0] in var3d) else ""} grid=False, **kw: obj[1].interp_xy(x, y,  {"z,"     if (obj[0] in var3d) else ""} grid, **kw)', dict(obj=obj)))

        #q.close()

        return


    def bathymetry_deriv(self, lat, lon, axis, grid=False):
        assert axis in ('lat', 'lon'), 'axis must be \'lat\' or \'lon\''
        return self.interps['bathymetry'].interp(lat, lon, grid, lat_deriv_order=(axis=='lat'), lon_deriv_order=(axis=='lon'))

    def bathymetry_deriv_xy(self, x, y, axis, grid=False):
        assert axis in ('x', 'y'), 'axis must be \'x\' or \'y\''
        return self.interps['bathymetry'].interp_xy(x, y, grid, x_deriv_order=(axis=='x'), y_deriv_order=(axis=='y'))

    def precip_type(self, lat, lon, epoch, grid=False):
        callback, varmap = precip_type_map[self.precip_src]
        v,y,x,t = callback(west=min(lon), east=max(lon), south=min(lat), north=max(lat), start=self.boundaries['start'], end=self.boundaries['end'])
        return np.array([varmap[v] for v in NearestNDInterpolator((y,x,t), v)(lat,lon,epoch)])

    def precip_type_xy(self, x, y, epoch, grid=False):
        callback, varmap = precip_type_map[self.precip_src]
        v,yy,xx,t = callback(west=min(x), east=max(x), south=min(y), north=max(y), start=self.boundaries['start'], end=self.boundaries['end'])
        return np.array([varmap[v] for v in NearestNDInterpolator((xx,yy,t), v)(x,y,epoch)])

    # legacy support
    def bathy(self, *args, **kwargs): return self.bathymetry(*args, **kwargs)
    def bathy_xy(self, *args, **kwargs): return self.bathymetry_xy(*args, **kwargs)
    def bathy_deriv(self, *args, **kwargs): return self.bathymetry_deriv(*args, **kwargs)
    def bathy_deriv_xy(self, *args, **kwargs): return self.bathymetry_deriv_xy(*args, **kwargs)
    def temp(self, *args, **kwargs): return self.temperature(*args, **kwargs)
    def temp_xy(self, *args, **kwargs): return self.temperature_xy(*args, **kwargs)



import os
import logging
from datetime import datetime, timedelta
from multiprocessing import Process, Queue

import numpy as np
import imageio
import matplotlib
#matplotlib.use('TkAgg')
#matplotlib.use('Qt5Agg')
import matplotlib.pyplot as plt
#import cartopy
#import cartopy.crs as ccrs
#import cartopy.feature as cfeature
from scipy.interpolate import griddata

from kadlu.geospatial.data_sources.source_map import load_map
from kadlu.geospatial.data_sources.data_util import fmt_coords
from kadlu.geospatial.data_sources.data_util import storage_cfg

'''

logging.getLogger('cartopy').setLevel(logging.WARNING)
logging.getLogger('matplotlib').setLevel(logging.WARNING)


config = dict(
        bgcontour   = lambda v: np.linspace(min(v)-.1, max(v)+.1, 3),
        bathymetry=dict(
            cm      = plt.cm.bone.reversed(),
            alpha   = 0.9,
            levels  = lambda v, n=12: np.linspace(1, max(v)-1, n),
            norm    = lambda v: matplotlib.colors.Normalize(vmin=0, vmax=max(v)-1),
            title   = 'bathymetry (metres)'),
        temp=dict(
            cm      = plt.cm.coolwarm, 
            alpha   = 0.8,
            levels  = lambda v, n=12: np.linspace(min(v)-.1, max(v)+.1, n),
            norm    = lambda v=None: matplotlib.colors.Normalize(vmin=-5, vmax=20),
            title   = 'temperature (celsius)'),
        salinity=dict(
            cm      = plt.cm.viridis,
            alpha   = 0.7,
            levels  = lambda v, n=12: np.linspace(min(v)+.1, max(v), n),
            norm    = lambda v=None: matplotlib.colors.Normalize(vmin=20, vmax=40),
            title   = 'salinity (g/kg salt in water)'),
        waveheight=dict(
            cm      = plt.cm.BuPu,
            alpha   = 0.85,
            levels  = lambda v, n=12: np.linspace(min(v)+.1, max(v), n),
            norm    = lambda v=None: matplotlib.colors.Normalize(vmin=0, vmax=15),
            title   = 'wave height (metres)')
    )


def plot2D(var, source, plot_wind=False, save=False, **kwargs): 
    """
    
        import kadlu
        var = 'bathymetry'
        source = 'gebco'
        plot_wind, save = False, False
        kwargs = kadlu.defaults
        kadlu.plot2D(var, source, **kwargs)

    """
    var = var.lower()
    source = source.lower()

    if f'{var}_{source}' not in load_map.keys():
        raise KeyError(f'could not find source for variable. valid vars and '
                       f'sources: {[k.split("_") for k in load_map.keys()]}')

    if 'start' not in kwargs.keys():
        kwargs['start'], kwargs['end'] = datetime.now(), datetime.now()

    data = load_map[f'{var}_{source}'](**kwargs)
    val, lat, lon = data[:3].astype(float)

    # project data onto coordinate space
    extent = ccrs.Mercator().transform_points(
            ccrs.Geodetic(),
            np.array([kwargs['west'], kwargs['east']]), 
            np.array([kwargs['south'], kwargs['north']])
        )[:,:-1]
    projected_lonlat = ccrs.Mercator().transform_points(
            ccrs.Geodetic(),
            lon,
            lat
        )
    plon = projected_lonlat[:,0]
    plat = projected_lonlat[:,1]
    num_lats = 1000
    num_lons = 1000
    lons = np.linspace(start=min(plon), stop=max(plon), num=num_lons)
    lats = np.linspace(start=min(plat), stop=max(plat), num=num_lats)
    data = griddata(points=(plon, plat), values=val, xi=(lons[None,:],lats[:,None]), method='linear')
    coast = cfeature.NaturalEarthFeature('physical', 'coastline', '10m')
    fg = (.92, .92, .92, 1)
    fname = f'{var}_{kwargs["start"].date().isoformat()}.png'
    fig = plt.figure()

    ax = fig.add_subplot(1, 1, 1, 
            title=config[var]['title']+f'\n{kwargs["start"].date().isoformat()}',
            projection=ccrs.Mercator(), 
            facecolor=config[var]['cm'](256), 
            frameon=True
        )
    ax.contourf(lons, lats, data,
                transform=ccrs.Mercator(),
                levels=config[var]['levels'](val),
                cmap=config[var]['cm'], 
                alpha=config[var]['alpha'],
                zorder=8
            )
    ax.contour(lons, lats, data,
                transform=ccrs.Mercator(),
                levels=config[var]['levels'](val),
                cmap=config[var]['cm'],
                alpha=1,
                linewidths=2,
                zorder=9
            )

    if plot_wind is not False:
        if plot_wind.lower() == 'era5': 
            #windfcnU, windfcnV = (Era5().load_wind_u, Era5().load_wind_v)
            windfcnU, windfcnV = load_map['wind_u_era5'], load_map['wind_v_era5']
        elif plot_wind.lower() == 'wwiii': 
            #windfcnU, windfcnV = (Wwiii().load_wind_u, Wwiii().load_wind_v)
            windfcnU, windfcnV = load_map['wind_u_wwiii'], load_map['wind_v_wwiii']
        else: 
            raise ValueError('invalid wind source. must be \'era5\' or \'wwiii\'')

        uval, ulat, ulon, utime = windfcnU(**kwargs)
        vval, vlat, vlon, vtime = windfcnV(**kwargs)
        assert(len(vval) == len(uval))  # this can be fixed with an SQL JOIN in load module
        if len(np.unique(ulat)) == 1 or len(np.unique(ulon)) == 1:
            raise RuntimeError(f'Not enough datapoints to plot windspeeds in region {fmt_coords(kwargs)}')

        ax.quiver(ulon, ulat, uval, vval, transform=ccrs.PlateCarree(), 
                regrid_shape=20, zorder=10)

    ax.add_feature(coast, facecolor=fg, edgecolor=(0,0,0,1), zorder=11)
    gl = ax.gridlines(crs=ccrs.PlateCarree(), draw_labels=True, linestyle='--',
            zorder=12)
    gl.xlabels_top = False
    gl.ylabels_right = False
    gl.xformatter = cartopy.mpl.gridliner.LONGITUDE_FORMATTER

    # this appears to be a known bug
    # https://github.com/SciTools/cartopy/issues/1332
    for tick in ax.get_xticklabels(): 
        tick.set_rotation(45)

    gl.yformatter = cartopy.mpl.gridliner.LATITUDE_FORMATTER
    ax.tick_params(axis='x', rotation=45)
    vnorm = val if var == 'bathy' or var == 'bathymetry' else None
    plt.colorbar(matplotlib.cm.ScalarMappable(norm=config[var]['norm'](vnorm),
                cmap=config[var]['cm']))

    if save is not False:
        if not os.path.isdir(f'{storage_cfg()}figures'): 
            os.mkdir(f'{storage_cfg()}figures')
        logging.info(f'saving figure to {storage_cfg()}figures/{fname if save is True else save}')
        plt.savefig(f'{storage_cfg()}figures/{fname if save is True else save}', 
                bbox_inches='tight', dpi=200, figsize=(12,8), optimize=True)
        plt.close()
    else: 
        plt.show()

    return


def animate(var, source, kwargs, step=timedelta(hours=12), fps=30, plot_wind=False, debug=False):
    """
    args:
        var: string
            view the list of available sources and variables with ``print(kadlu.source_map)``
        source: string
            view the list of available sources and variables with ``print(kadlu.source_map)``
        kwargs: dictionary
            dict containing boundary coordinates. example:
            ``kwargs=dict(south=44.25, west=-64.5, north=44.70, east=-63.33, top=0, bottom=5000, start=datetime(2015, 3, 1), end=datetime(2015, 3, 1, 12))``
        step: timedelta
            temporal resolution of individual video frames
        fps: int
            frames per second
        plot_wind: string
            if plot_wind is set to one of the available wind sources in kadlu.source_map, 
            wind direction markers will be overlayed on the map

    the resulting animation will be stored in the 'animated' subdirectory within the kadlu_data folder
    """

    # prepare folder and check for existing frames
    dirname = storage_cfg() + 'figures/'
    if not os.path.isdir(dirname): os.mkdir(dirname)
    png = lambda f: f if '.png' in f else None
    old = map(png, list(os.walk(dirname))[0][2])
    _rm = [os.remove(f'{dirname}{x}') for x in old if debug and x is not None]

    # generate image frames
    qry = kwargs.copy()
    cur = datetime(kwargs['start'].year, kwargs['start'].month, kwargs['start'].day)
    while cur <= kwargs['end']:
        qry['start'] = cur
        qry['end'] = cur + step
        fname = f'{var}_{cur.isoformat()}.png'
        if not os.path.isfile(f'{dirname}/{fname}'): 
            plot2D(var, source, plot_wind=plot_wind, save=fname, **qry)
        cur += step

    # filename and path for output
    fname = (f'{var}_{kwargs["start"].date().isoformat()}'
             f'_{kwargs["end"].date().isoformat()}.mp4')
    savedir = f'{storage_cfg()}animated{os.path.sep}'
    if not os.path.isdir(savedir): os.mkdir(savedir)

    # aggregate frames within query range and append to mp4 file
    logging.info(f'animating {fname}...')
    fmt = f'{var}_%Y-%m-%dT%H:%M:%S.png'
    frames = sorted([f'{dirname}{i}' for i in 
            map(png, list(os.walk(f'{dirname}'))[0][2]) if i is not None
            and datetime.strptime(i, fmt) >= kwargs['start']
            and datetime.strptime(i, fmt) <= kwargs['end']])
    with imageio.get_writer(f'{savedir}{fname}', mode='I', macro_block_size=4,
            format='FFMPEG', fps=fps) as w:
        list(map(w.append_data, map(imageio.imread, frames)))

    logging.info(f'saved animation to {savedir}{fname}')
    return 
'''


def plot_transm_loss_horiz(transm_loss, radial_axis, azimuthal_axis):
    """ Plot the transmission loss on a horizontal plane in polar coordinates.

        Args:
            transm_loss: numpy.array
                Transmission loss, has shape (nq,nr).
            radial_axis: numpy.array
                Radial axis, has shape (nr)
            azimuth_axis: numpy.array
                Azimuthal axis, has shape (nq)

        Returns:
            fig: matplotlib.figure.Figure
                A figure object.
    """
    # "complete the circle"
    azimuthal_axis = np.concatenate([azimuthal_axis, [np.pi]])
    transm_loss = np.concatenate([transm_loss, transm_loss[0:1,:]], axis=0)
    # convert to x,y meshgrid
    r, q = np.meshgrid(radial_axis, azimuthal_axis)
    x = r * np.cos(q) / 1e3
    y = r * np.sin(q) / 1e3
    # contour plot
    fig, ax = plt.subplots()
    img = ax.contourf(x, y, transm_loss, 100, cmap=matplotlib.cm.get_cmap('viridis_r'))
    # labels
    ax.set_xlabel('x (km)')
    ax.set_ylabel('y (km)')
    plt.title('Transmission loss')
    fig.colorbar(img, ax=ax, format='%2.0f dB')# colobar
    return fig


def plot_transm_loss_vert(transm_loss, vertical_axis, radial_axis, bathy_func=None, ssp_func=None):
    """ Plot the transmission loss on a vertical plane in carthesian coordinates.

        If a bathymetry interpolation function is provided, the seafloor will be 
        drawn superimposed on the transmission loss plot.

        Args:
            transm_loss: numpy.array
                Transmission loss, has shape (nz,nr).
            vertical_axis: numpy.array
                Vertical axis, has shape (nz)
             radial_axis: numpy.array
                Radial axis, has shape (nr)
            bathy_func: function
                Bathymetry interpolation function in radial variable, r
            ssp_func: function
                Sound speed interpolation function in radial variable, z

        Returnedme :
            fig: matplotlib.figure.Figure
                A figure object.
    """
    x, y = np.meshgrid(radial_axis, vertical_axis)
    # min and max transmission loss (excluding sea surface bin)
    tl_min = np.min(transm_loss[1:,:])
    tl_max = np.max(transm_loss[1:,:])
    # contour plot
    fig, ax = plt.subplots()
    img = ax.contourf(x/1e3, y, transm_loss, 100, vmin=tl_min, vmax=tl_max, cmap=matplotlib.cm.get_cmap('viridis_r'))
    ax.invert_yaxis()
    # labels
    ax.set_xlabel('r (km)')
    ax.set_ylabel('z (m)')
    plt.title('Transmission loss')
    fig.colorbar(img, ax=ax, format='%2.0f dB') # colobar
    # superimpose bathymetry
    if bathy_func is not None: 
        r_max = np.max(radial_axis)
        nr = min(10000, int(r_max / 10))
        r = np.linspace(0, r_max, num=nr)
        bathy = bathy_func(r)
        idx = np.argwhere(bathy <= vertical_axis[-1])
        ax.plot(r[idx]/1e3, bathy[idx], 'w')
    # superimpose ssp
    if ssp_func is not None: 
        z_max = np.max(np.max(vertical_axis))
        nz = min(1000, int(z_max))
        z = np.linspace(0, z_max, num=nz)
        ssp = ssp_func(z)
        axt = ax.twiny()
        dc = np.max(ssp) - np.min(ssp)
        axt.set_xlim(np.min(ssp)-0.2*dc, np.max(ssp)+0.2*dc)        
        axt.set_xlabel('c (m/s)')
        axt.plot(ssp, z, 'w', linestyle=':')

    return fig




""" Unit tests for the the 'sound.output' module in the 'kadlu' package

    Authors: Oliver Kirsebom
    contact: oliver.kirsebom@dal.ca
    Organization: MERIDIAN-Intitute for Big Data Analytics
    Team: Acoustic data Analytics, Dalhousie University
    Project: packages/kadlu
             Project goal: Tools for underwater soundscape modeling

    License:

"""
import pytest
import kadlu.sound.output.output_database as od
import os
import numpy as np
def test_create_dataset():
    # create sample dataset
    d = od.Dataset('sample.nc', 'w')
    d.close()

    # read the sample dataset into memory
    d = od.Dataset('sample.nc', 'r')

    # check if created test dataset exists
    filepath = d.filepath()
    assert os.path.isfile(filepath)

    # check dimensions attributes of sample dataset
    assert d.get_dim('X').size == 1061
    assert d.get_dim('Y').size == 711
    assert d.get_dim('SPL').size == 531
    assert d.get_dim('F').size == 1
    assert d.get_dim('nv').size == 2

    # check variables of sample dataset

    ## check variable 'spl'
    assert d.get_var('spl').getncattr('long_name') == "sound_pressure_level_in_water_threshold"
    assert d.get_var('spl').getncattr('coverage_content_type') == 'coordinate'
    assert d.get_var('spl').getncattr('axis') == 'SPL'
    assert d.get_var('spl').get_dims()[0].size == 531
    ## check variable 'y'
    assert d.get_var('y').ndim == 2
    assert d.get_var('y').dtype == np.double
    assert d.get_var('y').getncattr('units') == 'km'

    # check global attributes of sample dataset
    assert d.getncattr('naming_authority') == 'MERIDIAN'
    assert d.getncattr('history') == 'version 1.0'
    d.close()

    # add new variable into sample datasets
    d = od.Dataset('sample.nc', 'r+')
    d.new_var('test_var', 'f8', dimensions=('X', 'Y'), attrs={
        "property": {
            'standard_name': 'test_standard_name',
            'long_name': 'test_long_name',
            'units': 'km'
        }
    })
    d.set_global_attr('test_new_attr', 'add a new test global attribute')

    ## test newly added variable
    assert d.get_var('test_var').dtype == np.double
    assert d.get_var('test_var').ndim ==2
    assert d.get_var('test_var').getncattr('standard_name') == 'test_standard_name'

    ## test newly added global attribute
    assert d.get_global_attr('test_new_attr') == 'add a new test global attribute'
    d.close()

    # remove sample dataset after test
    os.remove(filepath)


"""
    API for Era5 dataset from Copernicus Climate Datastore
     
    Metadata regarding the dataset can be found here:
        https://cds.climate.copernicus.eu/cdsapp#!/dataset/reanalysis-era5-single-levels?tab=overview
"""

import os
import logging
import warnings
import configparser
from os.path import isfile, dirname
from datetime import datetime, timedelta

import cdsapi
import pygrib
import numpy as np

from kadlu import index
from kadlu.geospatial.data_sources.data_util    import (
        database_cfg,
        dev_null,
        dt_2_epoch,
        epoch_2_dt,
        logmsg,
        logmsg_nodata,
        storage_cfg,
        str_def,
    )


era5_tables  = [
        'significant_height_of_combined_wind_waves_and_swell',
        'mean_wave_direction',
        'mean_wave_period',
        'u_component_of_wind',
        'v_component_of_wind',
        'convective_precipitation',
        'convective_snowfall',
        'normalized_energy_flux_into_ocean',
        'normalized_energy_flux_into_waves',
        'normalized_stress_into_ocean',
        'precipitation_type',
    ]


logging.getLogger('cdsapi').setLevel(logging.WARNING)

def initdb():
    conn, db = database_cfg()
    for var in era5_tables:
        db.execute(f'CREATE TABLE IF NOT EXISTS {var}'
                    '( val     REAL    NOT NULL, ' 
                    '  lat     REAL    NOT NULL, ' 
                    '  lon     REAL    NOT NULL, ' 
                    '  time    INT     NOT NULL, ' 
                    '  source  TEXT    NOT NULL) ') 
        db.execute(f'CREATE UNIQUE INDEX IF NOT EXISTS '
                   f'idx_{var} on {var}(time, lon, lat, val, source)')
    conn.close()



def era5_cfg(key=None, url=None):

    cfg = configparser.ConfigParser()       # read .ini into dictionary object
    cfgfile = os.path.join(dirname(dirname(dirname(dirname(__file__)))), "config.ini")
    #cfgfile = os.path.join(storage_cfg(), "config.ini")
    cfg.read(cfgfile)

    if 'cdsapi' not in cfg.sections():
        cfg.add_section('cdsapi')
    if key is not None:
        cfg.set('cdsapi', 'key', key)
        with open(cfgfile, 'w') as f:
            cfg.write(f)
    else:
        cfg.set('cdsapi', 'key', '20822:2d1c1841-7d27-4f72-bb8a-9680a073b4c3')
        with open(cfgfile, 'w') as f:
            cfg.write(f)
    if url is not None:
        cfg.set('cdsapi', 'url', url)
        with open(cfgfile, 'w') as f:
            cfg.write(f)
    else:
        cfg.set('cdsapi', 'url', 'https://cds.climate.copernicus.eu/api/v2')
        with open(cfgfile, 'w') as f:
            cfg.write(f)
    return 


def fetch_era5(var, *, west, east, south, north, start, end, **_):
    """ fetch global era5 data for specified variable and time range

        args:
            var: string
                the variable short name of desired wave parameter 
                according to ERA5 docs.  the complete list can be found 
                here (table 7 for wave params):
                https://confluence.ecmwf.int/display/CKB/ERA5+data+documentation#ERA5datadocumentation-Temporalfrequency
            kwargs: dict
                keyword arguments passed from the Era5() class as a dictionary

        return:
            True if new data was fetched, else False 
    """
    # cleaner stack trace by raising outside of try/except
    err = False

    era5_cfg()
    cfg = configparser.ConfigParser()       # read .ini into dictionary object
    cfgfile = os.path.join(dirname(dirname(dirname(dirname(__file__)))), "config.ini")
    #cfgfile = os.path.join(storage_cfg(), "config.ini")
    cfg.read(cfgfile)

    try: c = cdsapi.Client(url=cfg['cdsapi']['url'], key=cfg['cdsapi']['key'])
    except KeyError:
        try: c = cdsapi.Client()
        except Exception:
            err = True  

    if err: 
        raise KeyError('CDS API has not been configured for the ERA5 module. '
                       'obtain an API token from the following URL and run '
                       'kadlu.era5_cfg(url="URL_HERE", key="TOKEN_HERE"). '
                       'https://cds.climate.copernicus.eu/api-how-to')

    t = datetime(start.year, start.month, start.day, start.hour)
    assert end - start <= timedelta(days=1, hours=1), 'use index to bin fetch requests'
        
    # fetch the data
    fname = f'ERA5_reanalysis_{var}_{t.strftime("%Y-%m-%d")}.grb2'
    fpath = f'{storage_cfg()}{fname}'
    if not isfile(fpath):
        with dev_null():
            c.retrieve('reanalysis-era5-single-levels', {
                       'product_type' : 'reanalysis',
                       'format'       : 'grib',
                       'variable'     : var,
                       'year'         : t.strftime("%Y"),
                       'month'        : t.strftime("%m"),
                       'day'          : t.strftime("%d"),
                       'time'         : [datetime(t.year, t.month, t.day, h)
                                         .strftime('%H:00') for h in range(24)]
                    }, fpath)
    
    # load the data file and insert it into the database
    assert isfile(fpath)
    grb = pygrib.open(fpath)
    agg = np.array([[],[],[],[],[]])
    table = var[4:] if var[0:4] == '10m_' else var

    for msg, num in zip(grb, range(1, grb.messages)):
        if msg.validDate < start or msg.validDate > end: 
            continue

        # read grib data
        z, y, x = msg.data()
        if np.ma.is_masked(z):
            z2 = z[~z.mask].data
            y2 = y[~z.mask]
            x2 = x[~z.mask]
        else:  # wind data has no mask
            z2 = z.reshape(-1)
            y2 = y.reshape(-1)
            x2 = x.reshape(-1)

        # adjust latitude-zero to 180th meridian
        x3 = ((x2 + 180) % 360) - 180

        # index coordinates, select query range subset, aggregate results
        xix = np.logical_and(x3>=west,  x3<=east)
        yix = np.logical_and(y2>=south, y2<=north)
        idx = np.logical_and(xix, yix)
        agg = np.hstack((agg, [z2[idx],
                               y2[idx],
                               x3[idx],
                               dt_2_epoch([msg.validDate for i in z2[idx]]),
                               ['era5' for i in z2[idx]]]))

    # perform the insertion
    #if 'lock' in kwargs.keys(): kwargs['lock'].acquire()
    initdb()
    conn, db = database_cfg()
    n1 = db.execute(f"SELECT COUNT(*) FROM {table}").fetchall()[0][0]
    db.executemany(f"INSERT OR IGNORE INTO {table} "
                   f"VALUES (?,?,?,CAST(? AS INT),?)", agg.T)
    n2 = db.execute(f"SELECT COUNT(*) FROM {table}").fetchall()[0][0]
    db.execute("COMMIT")
    conn.commit()
    conn.close()
    #insert_hash(kwargs, f'fetch_era5_{era5_varmap[var]}')
    #if 'lock' in kwargs.keys(): kwargs['lock'].release()

    logmsg('era5', var, (n1, n2), **dict(south=south, west=west, north=north, east=east, start=start, end=end))
    #logging.info(f"ERA5 {msg.validDate.date().isoformat()} {var}: "
    #             f"processed and inserted {n2-n1} rows in region "
    #             f"{dict(south=south, west=west, north=north, east=east, start=start, end=end)}. "
    #             f"{len(agg[0])- (n2-n1)} duplicates ignored")
    return True


def load_era5(var, *, west, east, south, north, start, end, **_):
    """ load era5 data from local database

        args:
            var:
                variable to be fetched (string)
            kwargs:
                dictionary containing the keyword arguments used for the
                fetch request. must contain south, north, west, east
                keys as float values, and start, end keys as datetimes

        return:
            values:
                values of the fetched var
            lat:
                y grid coordinates
            lon:
                x grid coordinates
            epoch:
                timestamps in epoch hours since jan 1 2000
    """
    # check for missing data
    with index(storagedir=storage_cfg(), west=west, east=east, south=south, north=north, start=start, end=end) as fetchmap: 
        fetchmap(callback=fetch_era5, var=var)

    # load the data
    table = var[4:] if var[0:4] == '10m_' else var  # table cant start with int
    sql = ' AND '.join([f"SELECT * FROM {table} WHERE lat >= ?",
        'lat <= ?',
        'lon >= ?',
        'lon <= ?',
        'time >= ?',
        'time <= ?']) + ' ORDER BY time, lat, lon ASC'
    conn, db = database_cfg()
    db.execute(sql, tuple(map(str, [
            south,                north, 
            west,                 east, 
            dt_2_epoch(start), dt_2_epoch(end)
        ])))
    rowdata = np.array(db.fetchall(), dtype=object).T
    conn.close()

    if len(rowdata) == 0:
        logmsg_nodata('era5', var, west=west, east=east, south=south, north=north, start=start, end=end)
        return np.array([[],[],[],[]])

    val, lat, lon, epoch, source = rowdata 
    return np.array((val, lat, lon, epoch), dtype=np.float)


class Era5():
    """ collection of module functions for fetching and loading  """

    def load_windwaveswellheight(self, **kwargs):
        return load_era5('significant_height_of_combined_wind_waves_and_swell', **kwargs)

    def load_wavedirection(self, **kwargs):
        return load_era5('mean_wave_direction', **kwargs)

    def load_waveperiod(self, **kwargs):
        return load_era5('mean_wave_period', **kwargs)

    def load_precipitation(self, **kwargs):
        return load_era5('convective_precipitation', **kwargs)

    def load_snowfall(self, **kwargs):
        return load_era5('convective_snowfall', **kwargs)

    def load_flux_ocean(self, **kwargs):
        return load_era5('normalized_energy_flux_into_ocean', **kwargs)

    def load_flux_waves(self, **kwargs):
        return load_era5('normalized_energy_flux_into_waves', **kwargs)

    def load_stress_ocean(self, **kwargs):
        return load_era5('normalized_stress_into_ocean', **kwargs)

    def load_precip_type(self, **kwargs):
        return load_era5('precipitation_type', **kwargs)

    def load_wind_u(self, **kwargs):
        return load_era5('10m_u_component_of_wind', **kwargs)

    def load_wind_v(self, **kwargs):
        return load_era5('10m_v_component_of_wind', **kwargs)

    def load_wind_uv(self, **kwargs):
        with index(storagedir=storage_cfg(), west=kwargs['west'], east=kwargs['east'], south=kwargs['south'], north=kwargs['north'], start=kwargs['start'], end=kwargs['end']) as fetchmap: 
            fetchmap(callback=fetch_era5, var='10m_u_component_of_wind')
            fetchmap(callback=fetch_era5, var='10m_v_component_of_wind')

        sql = ' AND '.join(['SELECT u_component_of_wind.val, u_component_of_wind.lat, u_component_of_wind.lon, u_component_of_wind.time, v_component_of_wind.val FROM u_component_of_wind '\
            'INNER JOIN v_component_of_wind '\
            'ON u_component_of_wind.lat == v_component_of_wind.lat',
            'u_component_of_wind.lon == v_component_of_wind.lon',
            'u_component_of_wind.time == v_component_of_wind.time '\
            'WHERE u_component_of_wind.lat >= ?',
            'u_component_of_wind.lat <= ?',
            'u_component_of_wind.lon >= ?',
            'u_component_of_wind.lon <= ?',
            'u_component_of_wind.time >= ?',
            'u_component_of_wind.time <= ?']) + ' ORDER BY u_component_of_wind.time, u_component_of_wind.lat, u_component_of_wind.lon ASC'
        conn, db = database_cfg()
        db.execute(sql, tuple(map(str, [
                kwargs['south'],                kwargs['north'], 
                kwargs['west'],                 kwargs['east'], 
                dt_2_epoch(kwargs['start']), dt_2_epoch(kwargs['end'])
            ])))

        wind_u, lat, lon, epoch, wind_v = np.array(db.fetchall()).T
        val = np.sqrt(np.square(wind_u) + np.square(wind_v))
        conn.close()
        return np.array((val, lat, lon, epoch)).astype(float)


    def __str__(self):
        info = '\n'.join([
                "Era5 Global Dataset from Copernicus Climate Datastore.",
                "Combines model data with observations from across",
                "the world into a globally complete and consistent dataset",
                "\thttps://cds.climate.copernicus.eu/cdsapp#!/dataset/reanalysis-era5-single-levels"])
        args = "(south, north, west, east, datetime, end)"
        return str_def(self, info, args)




import logging
from PIL import Image
from functools import reduce
from xml.etree import ElementTree as ET
import json
from datetime import datetime

import matplotlib
'''
matplotlib.use('tkagg')
'''
import mpl_scatter_density
import matplotlib.pyplot as plt
import netCDF4
import numpy as np

from kadlu.geospatial.data_sources.data_util import (
        database_cfg,
        dt_2_epoch,
        epoch_2_dt,
        index_arr,
        storage_cfg,
    )


def load_raster(filepath, plot=False, cmap=None, **kwargs):
    """ load data from raster file 

        args:
            filepath: string
                complete filepath descriptor of netcdf file to be read
            plot: boolean
                if True, a plot will be displayed using the qt5agg backend
            cmap: matplotlib colormap
                the full list of available colormaps can be viewed with:
                print(matplotlib.pyplot.colormaps())
                if None is supplied, pyplot will default to 
                matplotlib.pyplot.cm.cividis

        returns:
            values: numpy 2D array
            lats:   numpy 1D array
            lons:   numpy 1D array
    """
    if kwargs == {}: kwargs.update(dict(south=-90,west=-180,north=90,east=180))

    # load raster
    Image.MAX_IMAGE_PIXELS = 500000000
    #im = Image.open(filepath)

    with Image.open(filepath) as im:
        # GDAL raster format
        # http://duff.ess.washington.edu/data/raster/drg/docs/geotiff.txt
        if 33922 in im.tag.tagdata.keys():
            i,j,k,x,y,z = im.tag_v2[33922]  # ModelTiepointTag
            dx, dy, dz  = im.tag_v2[33550]  # ModelPixelScaleTag
            meta        = im.tag_v2[42112]  # GdalMetadata
            xml         = ET.fromstring(meta)
            params      = {tag.attrib['name'] : tag.text for tag in xml}
            lon = np.arange(x, x + (dx * im.size[0]), dx)
            lat = np.arange(y, y + (dy * im.size[1]), dy)[ :: -1] -90
            rng_lon = index_arr(kwargs['west'], lon), index_arr(kwargs['east'], lon)
            rng_lat = (abs(index_arr(kwargs['north'], lat[::-1])-len(lat)), 
                       abs(index_arr(kwargs['south'], lat[::-1])-len(lat)))
            logging.debug(f'{xml.tag}\nraster coordinate system: {im.tag_v2[34737]}'
                          f'\n{json.dumps(params, indent=2, sort_keys=True)}')

        # NASA / jet propulsion labs raster format (page 27)
        # https://landsat.usgs.gov/sites/default/files/documents/geotiff_spec.pdf
        elif 34264 in im.tag.tagdata.keys():
            dx,_,_,x,_,dy,_,y,_,_,dz,z,_,_,_,_ = im.tag_v2[34264]  # ModelTransformationTag
            lon = np.arange(x, x + (dx * im.size[0]), dx)
            lat = np.arange(y, y + (dy * im.size[1]), dy)
            rng_lon = index_arr(kwargs['west'], lon), index_arr(kwargs['east'], lon)
            rng_lat = index_arr(kwargs['south'], -lat),  index_arr(kwargs['north'], -lat)

        else: assert False, f'error {filepath}: unknown metadata tag encoding'
        assert not (z or dz), f'error {filepath}: 3D rasters not supported yet'

        aoi = im.crop((rng_lon[0], rng_lat[0], rng_lon[1], rng_lat[1])).load()
        grid = np.ndarray((len(range(*rng_lon)), len(range(*rng_lat))))

    n = reduce(np.multiply, (rng_lon[1] - rng_lon[0], rng_lat[1] - rng_lat[0]))
    if n > 10000000: 
        logging.info(f'this could take a few moments ({n} points)...')

    for xi in np.arange(rng_lon[0], rng_lon[1]) -rng_lon[0]:
        for yi in np.arange(rng_lat[0], rng_lat[1]) -rng_lat[0]: 
            grid[xi][yi] = aoi[int(xi),int(yi)]

    mask = grid == float(im.tag_v2[42113])
    val = np.ma.MaskedArray(grid, mask=mask)

    # plot the data
    if plot:
        x1, y1 = np.meshgrid(lon[rng_lon[0]:rng_lon[1]], lat[rng_lat[0]:rng_lat[1]], indexing='ij')
        fig = plt.figure()
        if (rng_lon[1]-rng_lon[0]) * (rng_lat[1]-rng_lat[0]) >= 100000:
            ax = fig.add_subplot(1,1,1, projection='scatter_density')
            plt.axis('scaled')
            raster = ax.scatter_density(x1, y1, c=val, cmap=cmap)
            plt.tight_layout()
        else:
            ax = fig.add_subplot(1,1,1)
            ax.scatter(x1, y1, c=val, cmap=cmap)
        plt.show()

    return val.data, lat[rng_lat[0]:rng_lat[1]], lon[rng_lon[0]:rng_lon[1]]


def load_netcdf(filename, var=None, gridXY=None, plot=False, cmap=matplotlib.pyplot.cm.bone.reversed(), **kwargs):
    """ read environmental data from netcdf and output to gridded numpy array

        args:
            filename: string
                complete filepath descriptor of netcdf file to be read
            var: string (optional)
                the netcdf attribute to be read as the values.
                by default, a guess will be made based on the file metadata
            gridXY: string
                netcdf filename describing grid values if lon/lat are stored as 
                grid index
            plot: boolean
                if True, a plot will be displayed using the qt5agg backend
            cmap: matplotlib colormap
                the full list of available colormaps can be viewed with:
                print(matplotlib.pyplot.colormaps())

        returns:
            values: numpy 2D array
            lats:   numpy 1D array
            lons:   numpy 1D array
    """
    if kwargs == {}: 
        kwargs.update(dict(south=-90,west=-180,north=90,east=180,
            start=datetime(1,1,1), end=datetime.now(), top=0, bottom=9999))
    ncfile = netCDF4.Dataset(filename)

    varmap = dict(
            MAPSTA      =   'f',  # ifremer tag: appears to be a land mask
            #hs             'v',
            lat         =   'y',
            latitude    =   'y',
            gridY       =   'y',
            lon         =   'x',
            longitude   =   'x',
            gridX       =   'x',
            time        =   't',
            epoch       =   't',
            depth       =   'z',
            elevation   =   'z',
        )

    axes = dict([(varmap[var][0], var) for var in ncfile.variables.keys() if var in varmap.keys()])
    uvars = [_ for _ in ncfile.variables.keys() if _ not in varmap.keys()]
    if len(uvars) > 0 or var: axes.update({'v': var or uvars[0]})
    if not var: var = 'z' if uvars == [] and var == None else uvars[0]

    assert 'x' in axes.keys(), f'missing x axis: {uvars}'
    assert 'y' in axes.keys(), f'missing y axis: {uvars}'
    assert len(axes) >= len(ncfile.variables.keys()), f'missing axis from: {uvars}'
    assert sum(key in varmap.keys() for key in ncfile.variables.keys()) >= len(axes)-1, 'not all vars match'
    assert len(uvars) <= 1, f'more than one unknown variable: {uvars}'

    logging.info(f'loading data from {ncfile.getncattr("title")}')

    if not gridXY:
        rng_lon = index_arr(kwargs['west'],  ncfile[axes['x']][:].data), index_arr(kwargs['east'],  ncfile[axes['x']][:].data)
        rng_lat = index_arr(kwargs['south'], ncfile[axes['y']][:].data), index_arr(kwargs['north'], ncfile[axes['y']][:].data)
    else:
        assert False
    '''
        ncgrid = netCDF4.Dataset(gridXY)
        rng_lat = index_arr(kwargs['south'],  ncgrid['latitude'][:].data.flatten()), index_arr(kwargs['north'],  ncgrid['latitude'][:].data.flatten())
        rng_lon = index_arr(kwargs['west'], ncfile[axes['x']][:].data), index_arr(kwargs['east'], ncfile[axes['x']][:].data)
    '''


    out = {} if not 'v' in axes.keys() else dict(val=ncfile[axes['v']][:].data[rng_lat[0]:rng_lat[1], rng_lon[0]:rng_lon[1]])
    out.update(dict(lat=ncfile[axes['y']][:].data[rng_lat[0]:rng_lat[1]], lon=ncfile[axes['x']][:].data[rng_lon[0]:rng_lon[1]]))

    # temporal index range
    if 't' in axes.keys(): 
        if ncfile.variables[axes['t']].units == 'days since 1990-01-01T00:00:00Z':
            t0 = datetime(1990,1,1)
            rng_t = (index_arr(dt_2_epoch(kwargs['start'], t0), ncfile[axes['t']][:].data * 24), 
                     index_arr(dt_2_epoch(kwargs['end'], t0), ncfile[axes['t']][:].data * 24))
            out['time'] = epoch_2_dt(ncfile[axes['t']][:].data[rng_t[0]:rng_t[1]] * 24, t0)
        if ncfile.variables[axes['t']].units == 'seconds since 1970-01-01T00:00:00Z':
            t0 = datetime(1970,1,1)
            rng_t = (index_arr(dt_2_epoch(kwargs['start'], t0), ncfile[axes['t']][:].data / 60 / 60 ), 
                     index_arr(dt_2_epoch(kwargs['end'], t0), ncfile[axes['t']][:].data / 60 / 60))
            out['time'] = epoch_2_dt(ncfile[axes['t']][:].data[rng_t[0]:rng_t[1]] , t0, unit='seconds')
        else:
            assert False, 'unknown time unit'

    # vertical index range
    if 'z' in axes.keys() and 'v' in axes.keys(): 
        assert axes['z'] != 'elevation', 'netcdf indexing error'
        rng_z = (index_arr(kwargs['top'], ncfile[axes['z']][:].data),
                 index_arr(kwargs['bottom'], ncfile[axes['z']][:].data))
        out['depth'] = ncfile[axes['z']][:].data[rng_z[0]:rng_z[1]]
    elif 'z' in axes.keys() and not 'v' in axes.keys() and len(axes.keys()) == 3:
        # when loading bathymetry, z-axis are the intended first column values
        out = dict(val=ncfile[axes['z']][:].data[rng_lat[0]:rng_lat[1], rng_lon[0]:rng_lon[1]], lat=out['lat'].copy(), lon=out['lon'].copy())
        if axes['z'] == 'elevation': out['val'] *= -1
    else: 
        assert 'v' in axes.keys(), 'something may have gone wrong here...'

    # plot the data
    if plot and len(out.keys()) == 3:
        x1, y1 = np.meshgrid(out['lon'], out['lat'], indexing='xy')
        fig = plt.figure()
        if (rng_lon[1]-rng_lon[0]) * (rng_lat[1]-rng_lat[0]) >= 100000:
            ax = fig.add_subplot(1,1,1, projection='scatter_density')
            plt.axis('scaled')
            pcm = ax.scatter_density(x1, y1, c=out['val'], cmap=cmap)
            plt.tight_layout()
        else:
            ax = fig.add_subplot(1,1,1)
            pcm = ax.scatter(x1, y1, marker='s', c=out['val'], cmap=cmap)
        fig.colorbar(pcm, ax=ax)
        plt.title('Depth (metres)')
        plt.show()

    lat_m, lon_m = np.meshgrid(out['lat'], out['lon'], indexing='ij')

    grid = np.vstack((out['val'].flatten().astype(float), lat_m.flatten().astype(float), lon_m.flatten().astype(float)))
    
    return grid


'''

kwargs = dict(south = 43.5841631194817, north = 44.01583688051829, west = -59.33904235799604, east = -58.740957642003956, top = 0, bottom = 10000)
filename = '/storage/GEBCO_2020.nc'
var=None
gridXY=None
plot = True
cmap = matplotlib.pyplot.cm.bone.reversed()

'''



"""
    data source:
        https://www.hycom.org/data/glbv1pt08

    web interface for manual hycom data retrieval:
        https://tds.hycom.org/thredds/dodsC/GLBv0.08/expt_53.X/data/2015.html
"""

import logging
import requests
import warnings
from functools import reduce
from datetime import datetime, timedelta

import numpy as np

from kadlu import index
from kadlu.geospatial.data_sources.data_util import (
        database_cfg,
        dt_2_epoch,
        epoch_2_dt,
        fmt_coords,
        index_arr,
        logmsg,
        logmsg_nodata,
        storage_cfg,
        str_def,
    )


#conn, db = database_cfg()
hycom_src = "https://tds.hycom.org/thredds/dodsC/GLBv0.08/expt_53.X/data"
hycom_tables = ['hycom_salinity', 'hycom_water_temp', 'hycom_water_u', 'hycom_water_v']


slices_str = lambda var, slices, steps=(1,1,1,1): f"{var}{''.join(map(lambda tup, step : f'[{tup[0]}:{step}:{tup[1]}]', slices, steps))}"

def initdb():
    conn, db = database_cfg()

    for var in hycom_tables:
        db.execute(f'CREATE TABLE IF NOT EXISTS {var}'
                    '( val     REAL NOT NULL,' 
                    '  lat     REAL NOT NULL,' 
                    '  lon     REAL NOT NULL,' 
                    '  time    INT  NOT NULL,' 
                    '  depth   INT  NOT NULL,' 
                    '  source  TEXT NOT NULL )')

        db.execute(f'CREATE UNIQUE INDEX IF NOT EXISTS '
                   f'idx_{var} on {var}(time, lon, lat, depth, val, source)')
    conn.close()

def fetch_grid(**_):
    """ download lat/lon/time/depth arrays for grid indexing.
        times are formatted as epoch hours since 2000-01-01 00:00 
    """
    import time
    logging.info("HYCOM  fetching coordinate index...")
    url = f"{hycom_src}/2015.ascii?lat%5B0:1:3250%5D,lon%5B0:1:4499%5D"
    grid_netcdf = requests.get(url)
    assert(grid_netcdf.status_code == 200)

    meta, data = grid_netcdf.text.split\
    ("---------------------------------------------\n")
    lat_csv, lon_csv = data.split("\n\n")[:-1]
    lat = np.array(lat_csv.split("\n")[1].split(", "), dtype=np.float)
    lon = np.array(lon_csv.split("\n")[1].split(", "), dtype=np.float)

    epoch = {}

    for year in map(str, range(1994, 2016)):
        logging.info(f"HYCOM  fetching time index for {year}...")
        url = f"{hycom_src}/{year}.ascii?time"
        time_netcdf = requests.get(url)
        assert(time_netcdf.status_code == 200)
        meta, data = time_netcdf.text.split\
        ("---------------------------------------------\n")
        csv = data.split("\n\n")[:-1][0]
        epoch[year] = np.array(csv.split("\n")[1].split(', ')[1:], dtype=float)
        time.sleep(1)

    logging.info(f"HYCOM  fetching depth index...")
    depth = np.array([0.0, 2.0, 4.0, 6.0, 8.0, 10.0, 12.0, 15.0, 20.0, 25.0,
        30.0, 35.0, 40.0, 45.0, 50.0, 60.0, 70.0, 80.0, 90.0, 100.0, 125.0,
        150.0, 200.0, 250.0, 300.0, 350.0, 400.0, 500.0, 600.0, 700.0, 800.0,
        900.0, 1000.0, 1250.0, 1500.0, 2000.0, 2500.0, 3000.0, 4000.0, 5000.0])

    return lat, lon, epoch, depth


def load_grid():
    """ put spatial grid into memory """
    with index(storagedir=storage_cfg(), bins=False, store=True) as fetchmap:
        return fetchmap(callback=fetch_grid, seed='hycom grids')[0]


class Hycom():
    """ collection of module functions for fetching and loading

        attributes:
            lat, lon: arrays
                spatial grid arrays. used to convert grid index to 
                coordinate point
            epoch: dict
                dictionary containing yearly time indexes.
                times formatted as epoch hours since 2000-01-01 00:00
            depth: array
                depth scale. used to convert depth index to metres
    """

    def __init__(self): 
        self.ygrid, self.xgrid, self.epoch, self.depth = None, None, None, None 

    def load_salinity (self, **kwargs): return self.load_hycom('salinity', kwargs)

    def load_temp     (self, **kwargs): return self.load_hycom('water_temp', kwargs)

    def load_water_u  (self, **kwargs): return self.load_hycom('water_u', kwargs)

    def load_water_v  (self, **kwargs): return self.load_hycom('water_v', kwargs)

    def callback(self, var, **kwargs): 
        """ build indices for query, fetch from hycom, insert into database
            
            args:
                var: string
                    variable to be fetched. complete list of variables here
                    https://tds.hycom.org/thredds/dodsC/GLBv0.08/expt_53.X/data/2015.html
                kwargs: dict
                    boundaries as keyword arguments
        """
        # build request indexes
        year = str(kwargs['start'].year)
        haystack = np.array([self.epoch[year], self.depth, self.ygrid, self.xgrid], dtype=object)
        needles1 = np.array([dt_2_epoch(kwargs['start']), kwargs['top'], kwargs['south'], kwargs['west'] ])
        needles2 = np.array([dt_2_epoch(kwargs['end']), kwargs['bottom'], kwargs['north'], kwargs['east'] ])
        slices = list(zip(map(index_arr, needles1, haystack), map(index_arr, needles2, haystack))) 
        #n = reduce(np.multiply, map(lambda s : s[1] - s[0] +1, slices))
        assert reduce(np.multiply, map(lambda s : s[1] - s[0] +1, slices)) > 0, "0 records available within query boundaries: {kwargs}"

        # generate request
        url = f"{hycom_src}/{year}.ascii?{slices_str(var, slices)}"
        with requests.get(url, stream=True) as payload_netcdf:
            assert payload_netcdf.status_code == 200, "couldn't access hycom server"
            meta, data = payload_netcdf.text.split\
            ("---------------------------------------------\n")

        # parse response into numpy array
        arrs = data.split("\n\n")[:-1]
        shape_str, payload = arrs[0].split("\n", 1)
        shape = tuple([int(x) for x in shape_str.split("[", 1)[1][:-1].split("][")])
        cube = np.ndarray(shape, dtype=np.float)

        for arr in payload.split("\n"):
            ix_str, row_csv = arr.split(", ", 1)
            a, b, c = [int(x) for x in ix_str[1:-1].split("][")]
            cube[a][b][c] = np.array(row_csv.split(", "), dtype=np.int)

        # build coordinate grid, populate with values, adjust scaling, remove nulls
        flatten = reduce(np.multiply, map(lambda s : s[1] - s[0] +1, slices))
        add_offset =  20 if 'salinity' in var or 'water_temp' in var else 0
        null_value = -10 if 'salinity' in var or 'water_temp' in var else -30
        grid = np.array([(None, y, x, t, d, 'hycom') 
                for t in self.epoch[year][slices[0][0] : slices[0][1] +1]
                for d in self.depth      [slices[1][0] : slices[1][1] +1]
                for y in self.ygrid      [slices[2][0] : slices[2][1] +1]
                for x in self.xgrid      [slices[3][0] : slices[3][1] +1]])
        grid[:,0] = np.reshape(cube, flatten) * 0.001 + add_offset
        grid = grid[grid[:,0] != null_value]

        # insert into db
        initdb()
        conn, db = database_cfg()
        n1 = db.execute(f"SELECT COUNT(*) FROM hycom_{var}").fetchall()[0][0]
        db.executemany(f"INSERT OR IGNORE INTO hycom_{var} VALUES "
                        "(?, ?, ?, CAST(? AS INT), CAST(? AS INT), ?)", grid)
        n2 = db.execute(f"SELECT COUNT(*) FROM hycom_{var}").fetchall()[0][0]
        db.execute("COMMIT")
        conn.commit()
        conn.close()

        logmsg('hycom', var, (n1, n2), **kwargs)
        return


    def fetch_hycom(self, var, kwargs): 
        """ convert user query to grid index slices, handle edge cases """
        assert kwargs['start'] <= kwargs['end']
        assert kwargs['start'] >  datetime(1994, 1, 1), 'data not available in this range'
        assert kwargs['end']   <  datetime(2016, 1, 1), 'data not available in this range'
        assert kwargs['south'] <= kwargs['north']
        assert kwargs['top']   <= kwargs['bottom']
        assert kwargs['start'] >= datetime(1994, 1, 1)
        assert kwargs['end']   <  datetime(2016, 1, 1)

        if self.ygrid is None:
            self.ygrid, self.xgrid, self.epoch, self.depth = load_grid()

        # if query spans antimeridian, make two seperate fetch requests
        if kwargs['west'] > kwargs['east']:
            logging.debug('splitting request')
            argsets = [kwargs.copy(), kwargs.copy()]
            argsets[0]['east'] = self.xgrid[-1]
            argsets[1]['west'] = self.xgrid[0]
        else: argsets = [kwargs]

        for argset in argsets:
            args = {k:v for k,v in argset.items() if k in ['west', 'east', 'south', 'north', 'top', 'bottom', 'start','end']}
            with index(storagedir=storage_cfg(), bins=True, **args) as fetchmap:
                fetchmap(callback=self.callback, var=var)

        return True


    def load_hycom(self, var, kwargs):
        """ load hycom data from local database

            args:
                var:
                    variable to be fetched. complete list of variables here
                    https://tds.hycom.org/thredds/dodsC/GLBv0.08/expt_53.X/data/2015.html
                south, north: float
                    ymin, ymax coordinate values. range: -90, 90
                kwargs: dict (boundaries as keyword arguments)
                    west, east: float
                        xmin, xmax coordinate values. range: -180, 180
                    start, end: datetime
                        temporal boundaries in datetime format

            return: 
                values: array
                    values of the fetched variable
                lat: array
                    y grid coordinates
                lon: array
                    x grid coordinates
                epoch: array
                    timestamps in epoch hours since jan 1 2000 
                depth: array
                    measured in meters
        """
        # check if grids are initialized
        if self.ygrid is None:
            self.ygrid, self.xgrid, self.epoch, self.depth = load_grid()

        # recursive function call for queries spanning antimeridian
        if (kwargs['west'] > kwargs['east']): 
            kwargs1 = kwargs.copy()
            kwargs2 = kwargs.copy()
            kwargs1['west'] = self.xgrid[0]
            kwargs2['east'] = self.xgrid[-1]
            return np.hstack((self.load_hycom(var, kwargs1), 
                              self.load_hycom(var, kwargs2)))

        # check for missing data
        self.fetch_hycom(var, kwargs)

        # validate and execute query
        assert 8 == sum(map(lambda kw: kw in kwargs.keys(),
                ['south', 'north', 'west', 'east',
                 'start', 'end', 'top', 'bottom'])), 'malformed query'

        assert kwargs['start'] <= kwargs['end']
        sql = ' AND '.join([f"SELECT * FROM hycom_{var} WHERE lat >= ?",
                'lat <= ?',
                'lon >= ?',
                'lon <= ?',
                'time >= ?',
                'time <= ?',
                'depth >= ?',
                'depth <= ?',
                "source == 'hycom' "]

            ) + 'ORDER BY time, depth, lat, lon ASC'

        conn, db = database_cfg()
        db.execute(sql, tuple(map(str, [
                kwargs['south'],                kwargs['north'], 
                kwargs['west'],                 kwargs['east'],
                dt_2_epoch(kwargs['start']),    dt_2_epoch(kwargs['end']),
                kwargs['top'],                  kwargs['bottom']
            ])))
        rowdata = np.array(db.fetchall(), dtype=object).T
        conn.close()

        if len(rowdata) == 0:
            #logging.warning(f'HYCOM {var.upper()}: no data found in region {fmt_coords(kwargs)}, returning empty arrays')
            logmsg_nodata('hycom', var, **kwargs)
            return np.array([[],[],[],[],[]])

        return rowdata[0:5].astype(float)


    def load_water_uv (self, **kwargs):
        self.fetch_hycom('water_u', kwargs)
        self.fetch_hycom('water_v', kwargs)

        sql = ' AND '.join(['SELECT hycom_water_u.val, hycom_water_u.lat, hycom_water_u.lon, hycom_water_u.time, hycom_water_u.depth, hycom_water_v.val FROM hycom_water_u '\
                'INNER JOIN hycom_water_v '\
                'ON hycom_water_u.lat == hycom_water_v.lat',
                'hycom_water_u.lon == hycom_water_v.lon',
                'hycom_water_u.time == hycom_water_v.time '\
                'WHERE hycom_water_u.lat >= ?',
                'hycom_water_u.lat <= ?',
                'hycom_water_u.lon >= ?',
                'hycom_water_u.lon <= ?',
                'hycom_water_u.time >= ?',
                'hycom_water_u.time <= ?',
                'hycom_water_u.depth >= ?',
                'hycom_water_u.depth <= ?',
            ]) + ' ORDER BY hycom_water_u.time, hycom_water_u.lat, hycom_water_u.lon ASC'

        conn, db = database_cfg()
        db.execute(sql, tuple(map(str, [
                kwargs['south'],                kwargs['north'], 
                kwargs['west'],                 kwargs['east'], 
                dt_2_epoch(kwargs['start']),    dt_2_epoch(kwargs['end']),
                kwargs['top'],                  kwargs['bottom'],
            ])))
        qry = np.array(db.fetchall()).T
        conn.close()

        if len(qry) == 0:
            logging.warning(f'HYCOM water_uv: no data found in region {fmt_coords(kwargs)}, returning empty arrays')
            return np.array([[],[],[],[],[]])

        water_u, lat, lon, epoch, depth, water_v = qry
        val = np.sqrt(np.square(water_u) + np.square(water_v))
        return np.array((val, lat, lon, epoch, depth)).astype(float)


    def __str__(self):
        info = '\n'.join([
                "Native hycom .[ab] data converted to NetCDF at the Naval",
                "Research Laboratory, interpolated to 0.08° grid between",
                "40°S-40°N (0.04° poleward) containing 40 z-levels.",
                "Availability: 1994 to 2015",
                "\thttps://www.hycom.org/data/glbv0pt08" ])

        args = ("(south, north, west, east, "
                "start, end, top, bottom)")

        return str_def(self, info, args)
    




import os
import sys
import logging
import requests
from ftplib import FTP

import kadlu

def fetch_netcdf_FTP(ftpurl, ftpdir, localdir):
    """ fetch files from an FTP server

        args:
            ftpurl: string
                FTP domain to connect
            ftpdir: string
                collect files within this FTP path directory
            localdir: string
                save files to this location
    """
    ftp = FTP(ftpurl)
    ftp.login()
    ftp.cwd(ftpdir)
    with kadlu.Capturing() as output: ftp.retrlines('NLST')
    for ncfile in [f for f in output if kadlu.ext(f, ('.nc',))]:
        logging.info(f'fetching {ncfile}')
        with open(f'{localdir}{ncfile}', 'wb') as fp:
            ftp.retrbinary(f'RETR {ncfile}', fp.write)


class Ifremer():
    """ collect files from ifremer FTP server """

    def fetch_ifremer_netcdf_hs2013(self, localdir=f'{kadlu.storage_cfg()}testfiles/'): 
        """ download netcdf waveheight files for testing kadlu """
        if not os.path.isdir(localdir): os.mkdir(localdir)
        fetch_netcdf_FTP(ftpurl='ftp.ifremer.fr', ftpdir='/ifremer/ww3/HINDCAST/ATNW/2013_CFSR/hs/', localdir=localdir)





import os 
import subprocess

import kadlu

class test_files():
    """ start a debugging session for reading .nc and .tif file formats.

    dynamically generates tests for a given set of input data files.
    pytest tests will be written to a script using files in 
    ``kadlu_data/testfiles/`` as inputs. 
    this is then run in a pytest subprocess, the test script will be cleaned 
    upon exit.

    pytest config can be passed using the ``DEBUGOPTS`` environment variable:
    by default the script will drop into an interactive pdb debugging session.
    to set a breakpoint, add ``breakpoint()`` directly to the source code.
    otherwise, pdb will break on exceptions.

    alternatively, run tests in parallel using the pytest-parallel package

    Usage:

    .. code-block:: python

        from importlib import reload
        import os, kadlu

        # place some files in the testfiles directory
        kadlu.ifremer().fetch_ifremer_netcdf_hs2013()
        
        # reload kadlu and run tests
        reload(kadlu); kadlu.test_files()
        
        # optionally pass args to pytest with DEBUGOPTS
        os.environ.setdefault('DEBUGOPTS', '--pdb')  # default behaviour
        reload(kadlu); kadlu.test_files()
    
    From the terminal:

    .. code-block:: bash

        # run pytest with pdb, shorter traceback, and program output
        export DEBUGOPTS='--pdb --tb=short -s'
        python3 test_files.py

        # run tests in parallel and log the results
        pip install pytest-parallel
        export DEBUGOPTS='--workers=auto --tb=line --durations=0' 
        python3 -B test_files.py | tee testresults.log

    see the ``DEBUGOPTS`` env var and 'man pytest' for further usage information
    """

    def __init__(self):
        with self as tests: tests.run()

    def __enter__(self):
        IMPORTS = 'import os, kadlu'
        PATH, _, FILES = list(os.walk(kadlu.storage_cfg()+'testfiles'))[0]
        TESTS = lambda F,PATH=PATH: F'''

    def test_loadfile_{F.replace('.','').replace('-','').replace(' ','')}():
        kadlu.load_file(os.path.join('{PATH}','{F}')), 'error: {F}' '''
        with open('scriptoutput.py', 'w') as OUTPUT: OUTPUT.write(IMPORTS+'\nif 21392>>4:'+''.join(map(TESTS, sorted(FILES))))
        return self

    def run(self):
        subprocess.run(f'python3.8 -B -m pytest {os.environ.get("DEBUGOPTS", "--pdb --tb=native -s")} scriptoutput.py'.split())

    def __exit__(self, exc_type, exc_value, traceback):
        os.remove('scriptoutput.py')


if __name__ == '__main__':
    test_files()




import pytest
import kadlu
import numpy as np
from datetime import datetime, timedelta
import os


# gulf st lawrence - small test area
south =  45.01
north =  45.99923
west  = -63.99
east  = -63
top   =  0
bottom=  5000
start = datetime(2000, 1, 10)
end   = datetime(2000, 1, 10, 12)

bounds = dict(
        south =  45,
        north =  46,
        west  = -64,
        east  = -63,
        top   =  0,
        bottom=  5000,
        start = datetime(2000, 1, 10),
        end   = datetime(2000, 1, 10, 12),
    )

def test_load_salinity():
    val, lat, lon, time, depth = kadlu.load(var='salinity', source='hycom', south=south, north=north, west=west, east=east, start=start, end=end, top=top, bottom=bottom)
    assert(len(val) == len(lat) == len(lon) == len(time))
    assert(sum(lat <= 90) == sum(lat >= -90) == len(lat))
    assert(sum(lon <= 180) == sum(lon >= -180) == len(lon))

def test_fetch_load_over_antimeridian():
    south, west = 44, 179
    north, east = 45, -179
    top, bottom = 0, 100

    val, lat, lon, time, depth = kadlu.load(var='salinity', source='hycom',
            south=south, north=north, 
            west=west, east=east, 
            start=start, end=end, 
            top=top, bottom=bottom
        )
    
    # commented to improve test speed
    """
    assert(len(val) > 0)
    assert(len(val) == len(lat) == len(lon) == len(time))
    assert(sum(lat <= 90) == sum(lat >= -90) == len(lat))
    assert(sum(lon <= 180) == sum(lon >= -180) == len(lon))

    assert np.all(lat >= south)
    assert np.all(lat <= north)
    #assert np.all(lon >= east)
    #assert np.all(lon <= west)
    """

# matt_s 2019-12
# hycom connection seems to be pretty slow for some reason... im getting ~2kbps download speeds
# in the meantime i've commented out the other fetch tests to make integrated testing faster

#def test_fetch_temp():
#    hycom.Hycom().fetch_temp(south=south, north=north, west=west, east=east, start=start, end=end, top=top, bottom=bottom)

def test_load_temp():
    val, lat, lon, time, depth = kadlu.load(source='hycom', var='temp', south=south, north=north, west=west, east=east, start=start, end=end, top=top, bottom=bottom)
    assert(len(val) == len(lat) == len(lon) == len(time))
    assert(sum(lat <= 90) == sum(lat >= -90) == len(lat))
    assert(sum(lon <= 180) == sum(lon >= -180) == len(lon))

#def test_fetch_water_u():
#    hycom.Hycom().fetch_water_u(south=south, north=north, west=west, east=east, start=start, end=end, top=top, bottom=bottom)

def test_load_water_uv():
    bounds = dict(
            south =  44.01,
            north =  44.31,
            west  = -63.1,
            east  = -62.9,
            top   =  0,
            bottom=  500,
            start = datetime(2000, 1, 10),
            end   = datetime(2000, 1, 10, 12),
        )

    data = kadlu.load(source='hycom', var='water_uv', **bounds)

def test_load_water_u():
    uval = kadlu.load(source='hycom',var='water_u', **bounds)
    # commented to improve test speed
    """
    assert(len(val) == len(lat) == len(lon) == len(time))
    assert(sum(lat <= 90) == sum(lat >= -90) == len(lat))
    assert(sum(lon <= 180) == sum(lon >= -180) == len(lon))
    """

#def test_fetch_water_v():
#    hycom.Hycom().fetch_water_v(south=south, north=north, west=west, east=east, start=start, end=end, top=top, bottom=bottom)

def test_load_water_v():
    vval = kadlu.load(source='hycom',var='water_v', **bounds)
    # commented to improve test speed
    """
    assert(len(val) == len(lat) == len(lon) == len(time))
    assert(sum(lat <= 90) == sum(lat >= -90) == len(lat))
    assert(sum(lon <= 180) == sum(lon >= -180) == len(lon))
    """


""" interactive mode debugging: assert db ordering is correct

    step through fetch_hycom() and put output and grid arrays into memory. 
    example test input:
>>>     
        year = '2000'
        var = 'salinity'
        slices = [
            (0, 2),         # time: start, end 
            (0, 3),         # depth: top, bottom
            (800, 840),     # x grid index: lon min, lon max
            (900, 1000)     # y grid index: lat min, lat max
        ]
        lat, lon = load_grid()
        epoch = load_times()
        depth = load_depth()

    run through the output builder loop again. this time, add an assertion to check 
    that the 4D array was flattened correctly
>>>     
        ix = 0  # debug index: assert order is correct
        for arr in payload.split("\n"):
            ix_str, row_csv = arr.split(", ", 1)
            a, b, c = [int(x) for x in ix_str[1:-1].split("][")]
            # output[a][b][c] = np.array(row_csv.split(", "), dtype=np.int)
            assert((output[a][b][c] == grid[ix:ix+len(output[a][b][c]), 0]).all())
            ix += len(output[a][b][c])

"""
"""
    import timeit 

    lat, sorted_arr = load_grid()
    def fcn():
        return index(np.random.rand()*360 - 180, sorted_arr)
    
    timeit.timeit(fcn, number=1000000)
    
"""


"""
    var = 'salinity'
    qry = {
            'south':44,  
            'north':45,
            'west':-60,
            'east':-59,
            'top':0,
            'bottom':5000,
            'start':datetime(2015, 10, 1),
            'end':datetime(2015, 10, 1, 12)
        }
    self = hycom.Hycom()
"""



"""
    data fetching utils, function maps, and constant variables
    collection of objects
"""

from datetime import datetime, timedelta

#import kadlu.geospatial.data_sources.chs as chs
import kadlu.geospatial.data_sources.era5 as era5
import kadlu.geospatial.data_sources.gebco as gebco
import kadlu.geospatial.data_sources.hycom as hycom
import kadlu.geospatial.data_sources.wwiii as wwiii
#import kadlu.geospatial.data_sources.nemo as nemo


# dict for mapping strings to callback functions
load_map = dict(
#        bathy_chs           = chs  .Chs()  .load_bathymetry,
        #temp_hycom          = hycom.Hycom().load_temp,
        temperature_hycom   = hycom.Hycom().load_temp,
        salinity_hycom      = hycom.Hycom().load_salinity,
        water_uv_hycom      = hycom.Hycom().load_water_uv,
        water_u_hycom       = hycom.Hycom().load_water_u,
        water_v_hycom       = hycom.Hycom().load_water_v,
        wavedir_era5        = era5 .Era5() .load_wavedirection,
        wavedirection_era5  = era5 .Era5() .load_wavedirection,
        waveheight_era5     = era5 .Era5() .load_windwaveswellheight,
        waveperiod_era5     = era5 .Era5() .load_waveperiod,
        wind_uv_era5        = era5 .Era5() .load_wind_uv,
        wind_u_era5         = era5 .Era5() .load_wind_u,
        wind_v_era5         = era5 .Era5() .load_wind_v,
        precipitation_era5  = era5 .Era5() .load_precipitation,
        snowfall_era5       = era5 .Era5() .load_snowfall,
        flux_ocean_era5     = era5 .Era5() .load_flux_ocean,
        flux_waves_era5     = era5 .Era5() .load_flux_waves,
        stress_ocean_era5   = era5 .Era5() .load_stress_ocean,
        #precip_type_era5    = era5 .Era5() .load_precip_type,
        wavedir_wwiii       = wwiii.Wwiii().load_wavedirection,
        wavedirection_wwiii = wwiii.Wwiii().load_wavedirection,
        waveheight_wwiii    = wwiii.Wwiii().load_windwaveheight,
        waveperiod_wwiii    = wwiii.Wwiii().load_waveperiod,
        wind_uv_wwiii       = wwiii.Wwiii().load_wind_uv,
        wind_u_wwiii        = wwiii.Wwiii().load_wind_u,
        wind_v_wwiii        = wwiii.Wwiii().load_wind_v,
        #bathy_gebco         = gebco.Gebco().load_bathymetry,
        bathymetry_gebco    = gebco.Gebco().load_bathymetry,
        #water_u_nemo        = nemo .Nemo().load_uvelocity,
        #water_v_nemo        = nemo .Nemo().load_vvelocity,
        #water_w_nemo        = nemo .Nemo().load_wvelocity,
        #salinity_nemo       = nemo .Nemo().load_salinity,
        #temp_nemo           = nemo .Nemo().load_temperature,
        #turbidity_nemo      = nemo .Nemo().load_fraserturbidity,
    )

precip_type_era5_map = {
        0   :   'Reserved',
        1   :   'Rain',
        2   :   'Thunderstorm',
        3   :   'Freezing rain',
        4   :   'Mixed/Ice',
        5   :   'Snow',
        6   :   'Wet snow',
        7   :   'Mixture of rain and snow',
        8   :   'Ice pellets',
        9   :   'Graupel',
        10  :   'Hail',
        11  :   'Drizzle',
        12  :   'Freezing drizzle',
        255 :   'Missing',
    }
_ = [precip_type_era5_map.update({i : 'Reserved'}) for i in range(13, 192)]
_ = [precip_type_era5_map.update({i : 'Reserved for local use'}) for i in range(192, 255)]

precip_type_map = dict(
    era5 = (era5.Era5().load_precip_type, precip_type_era5_map)
)

default_val = dict(
        south=44.25, west=-64.5,
        north=44.70, east=-63.33,
        top=0, bottom=5000,
        start=datetime(2015, 3, 1), end=datetime(2015, 3, 1, 12)
    )

var3d = ('temp', 'temperature', 'salinity', 'water_u', 'water_v', 'water_uv',)

source_map = (
    """
    CHS   (Canadian Hydrography Service)
          bathymetry:       bathymetric data in Canada's waterways. metres, variable resolution \n
    ERA5  (Global environmental dataset from Copernicus Climate Data Store)
          wavedir:          mean wave direction, degrees
          waveheight:       combined height of wind, waves, and swell. metres
          waveperiod:       mean wave period, seconds
          wind_uv:          wind speed computed as sqrt(u^2 + v^2) / 2, where u, v are direction vectors
          wind_u:           wind speed coordinate U-vector, m/s
          wind_v:           wind speed coordinate V-vector, m/s
          precipitation:
          precip_type:
          snowfall:
          flux_ocean:
          flux_waves:
          stress_ocean: \n
    GEBCO (General Bathymetric Chart of the Oceans)
          bathymetry:       global bathymetric and topographic data. metres below sea level \n
    HYCOM (Hybrid Coordinate Ocean Model)
          salinity:         g/kg salt in water
          temp:             degrees celsius
          water_uv:         ocean current computed as sqrt(u^2 + v^2) / 2, where u, v are direction vectors
          water_u:          ocean current coordinate U-vector, m/s
          water_v:          ocean current coordinate V-vector, m/s \n
    WWIII (WaveWatch Ocean Model Gen 3)
          wavedir:          primary wave direction, degrees
          waveheight:       combined height of wind and waves, metres
          waveperiod:       primary mean wave period, seconds
          wind_uv:          wind speed computed as sqrt(u^2 + v^2) / 2, where u, v are direction vectors
          wind_u:           wind speed coordinate U-vector, m/s
          wind_v:           wind speed coordinate V-vector, m/s \n
    NEMO  (3D zonal component fields from the SalishSeaCoast Model)
          water_u:
          water_v:
          water_w:
          salinity:
          temp:
          turbidity:
          mesozooplankton:
          microzooplankton:

    """)


class SourceMap():
    def fetch_map(self): return fetch_map 
    def load_map(self): return load_map
    def default_val(self): return default_val
    def var3d(self): return var3d
    def source_map(self): return source_map 




import os
import logging

LOGLEVEL = os.environ.get('LOGLEVEL', 'INFO')
#logging.basicConfig(format='%(asctime)s  %(message)s', level=LOGLEVEL, datefmt='%Y-%m-%d %I:%M:%S')
logging.basicConfig(format='%(message)s', level=LOGLEVEL, datefmt='%Y-%m-%d %I:%M:%S')

# job scheduler 
# used for automatic fetching
# must be loaded first to avoid circular dependency error
from .index import index, parallelindex

# data utils
from .geospatial.data_sources.data_util import (
        Capturing,
        database_cfg,
        dt_2_epoch,
        epoch_2_dt,
        ext,
        index_arr,
        reshape_2D,
        reshape_3D,
        storage_cfg,
        verbosity,
    )
from .geospatial.data_sources.source_map import source_map 
from .geospatial.data_sources.source_map import default_val as defaults
from .geospatial.data_sources.load_from_file import load_netcdf, load_raster

# data sources
from .geospatial.data_sources.era5 import Era5 as era5, era5_cfg
from .geospatial.data_sources.gebco import Gebco as gebco
from .geospatial.data_sources.hycom import Hycom as hycom
from .geospatial.data_sources.wwiii import Wwiii as wwiii
from .geospatial.data_sources.ifremer import Ifremer as ifremer 
from .geospatial.data_sources.source_map import load_map

# acoustic and environmental processing
from .geospatial.ocean import Ocean
from .sound.geophony import *
from .sound.parabolic_equation import *
from .sound.sound_speed import *


# file testing for files in kadlu_data/testfiles/
from .tests.geospatial.data_sources.test_files import test_files

# plotting tools
from .plot_util import (
#        animate,
#        plot2D,
        plot_transm_loss_horiz,
        plot_transm_loss_vert,
    )



def load(source, var, **kwargs):
    """ automated fetching and loading from web sources 

        args
            source: string
                view the list of available sources and variables with ``print(kadlu.source_map)``
            var: string
                view the list of available sources and variables with ``print(kadlu.source_map)``
            kwargs: dictionary
                dict containing boundary coordinates. example:
                ``kwargs=dict(south=44.25, west=-64.5, north=44.70, east=-63.33, top=0, bottom=5000, start=datetime(2015, 3, 1), end=datetime(2015, 3, 1, 12))``

        returns: np.ndarray 
            numpy arrays containing data ordered by val, lat, lon, [time, depth].
            times are in epoch format
    """

    source, var = source.lower(), var.lower()
    if var == 'bathy' or var == 'depth' or var == 'elevation': var = 'bathymetry'
    if var == 'temp': var = 'temperature'

    loadkey = f'{var}_{source}'
    assert loadkey in load_map.keys(), f'error: invalid source or variable. valid options include: \n\n'\
            f'{list(f.rsplit("_", 1)[::-1] for f in load_map.keys())}\n\n'\
            f'for more info, print(kadlu.source_map)'

    return load_map[loadkey](**kwargs)


def load_file(filepath, **kwargs):
    """ loading from local files. currently supports geotiff raster formats as well as certain netcdf databases

        args
            filepath: string
                full path directory and filename of data to load

            kwargs: dictionary
                dict containing coordinate bounds, e.g. north, south, 
                east, west, top, bottom, start, end
                note that times are in datetime format

        returns an ND numpy array. arrays ordered by:
            val, lat, lon, [time, depth]
            times are in epoch format
    """
    assert os.path.isfile(filepath), f'error: could not find {filepath}'

    if ext(filepath, ('.nc',)):
        return load_netcdf(filepath, **kwargs)

    elif ext(filepath, ('.tif', '.tiff', '.gtiff',)):       
        return load_raster(filepath, **kwargs)

    else:
        assert False, f'error {filepath}: unknown file format - currently only .nc and .tif formats are accepted'




""" Unit tests for the sound speed module within the kadlu library
"""
import pytest
import os
import numpy as np
from kadlu.sound.sound_speed import SoundSpeed
from kadlu.geospatial.ocean import Ocean

path_to_assets = os.path.join(os.path.dirname(os.path.dirname(__file__)),"assets")


def test_ocean_or_ssp_must_be_specified():
    with pytest.raises(AssertionError):
        SoundSpeed(num_depths=50, rel_err=None)

def test_sound_speed_from_uniform_data():
    # environment data provider
    o = Ocean(load_bathymetry=1000, load_temp=4, load_salinity=3)
    # instance of sound speed class 
    _ = SoundSpeed(o, num_depths=50, rel_err=None)

def test_sound_speed_from_uniform_ssp():
    # instance of sound speed class 
    ss = SoundSpeed(ssp=1499, num_depths=50, rel_err=None)
    # evaluate
    c = ss.interp_xy(x=0, y=0, z=0)
    assert c == 1499

def test_sound_speed_from_ssp():
    # sound speed profile
    z0 = np.array([0, 10, 20, 30, 60])
    c0 = np.array([1500, 1510, 1512, 1599, 1489])
    # instance of sound speed class 
    ss = SoundSpeed(ssp=(c0,z0), num_depths=50, rel_err=None)
    # evaluate
    c = ss.interp_xy(x=0, y=0, z=z0, grid=True)
    assert np.all(np.abs(c-c0) < 1E-6)


import os
import json
import pickle
import sqlite3
import logging
from hashlib import md5
from multiprocessing import Pool
from datetime import datetime, timedelta
from collections.abc import Iterable

from numpy import arange


class index():
    """ job scheduler utility featuring:
            - spatial hashing, indexing, and binning for spatial processing
            - in-memory caching for arbitrary binary objects
            - parallel process management (see inherited: parallelindex)
    
        intended to be a lightweight utility for hashmap-like boolean 
        indexing on function calls, performant storage of in-memory binary 
        objects mapped to boundary regions, and parallel process scheduling to 
        run spatial/temporal processing in smaller batches

        for usage: see __main__ demo at bottom of script
        requires python3.8
    """

    # compute 8-bit integer hash for a given dictionary
    hash_dict = lambda self, kwargs, seed='': int(md5((str(seed) + json.dumps(kwargs, sort_keys=True, default=str)).encode('utf-8')).hexdigest(), base=16) >> 80

    def __init__(self, *, pool=1, store=False, inmemory=False, bins=True, dx=2, dy=2, dz=5000, dt=timedelta(days=1), storagedir=os.getcwd(), filename='checksums.db', **kwargs): 
        """
            args:
                pool:
                    number of processes to run in parallel. when not using the
                    parallelindex subclass, this must be equal to 1.
                store:
                    determines whether to store the results of the callback.
                    when False, no binary will be serialized. only a hash of the 
                    function and input arguments will be stored to determine
                    whether the callback was called
                inmemory:
                    when store=True and inmemory=True, callback results can be 
                    stored to an in-memory database. this allows for faster
                    read and update times for processing that requires frequent
                    updates. will be copied to disk upon __exit__
                bins:
                    if True, boundary kwargs supplied to __init__ will be 
                    split into multiple smaller area bins and passed to the 
                    callback function. when 
                dx:
                    delta longitude bin size (float)
                dy: 
                    delta latitude bin size (float)
                dz:
                    delta depth bin size (float)
                dt:
                    delta time bin size (timedelta)
                storagedir:
                    directory to store spatial hashes and binary objects
               **kwargs:
                    boundary arguments to split into bins, e.g. 
                    (south, west, north, east, top, bottom, start, end)
        """
        assert kwargs != {} or bins == False, 'no boundaries provided'
        self.store, self.pool, self.storagedir, self.inmemory, = store, pool, storagedir, inmemory
        self.storage = os.path.join(storagedir, filename) if not inmemory else ':memory:'
        self.kwargslist = list(self.bin_kwargs(dx, dy, dz, dt, **kwargs)) if bins else [kwargs]

    def __enter__(self):
        assert self.kwargslist != [], 'empty kwargs!'
        assert os.path.isdir(str(self.storagedir)), f'invalid dir {storagedir}'
        with sqlite3.connect(self.storage) as con:
            db = con.cursor()
            db.execute('CREATE TABLE IF NOT EXISTS hashmap(hash INT NOT NULL, bytes BLOB)')
            db.execute('CREATE UNIQUE INDEX IF NOT EXISTS idx_map on hashmap(hash)')
        # TODO: 
        # if self.inmemory: copy disk database into memory
        # https://stackoverflow.com/questions/4019081/how-to-copy-a-sqlite-table-from-a-disk-database-to-a-memory-database-in-python
        assert not self.inmemory, 'feature not yet implemented'  
        return self

    def __call__(self, *, callback,  **passkwargs):  # https://docs.python.org/3/whatsnew/3.8.html
        return list(self.__call_generator__(callback=callback, **passkwargs))

    def __call_generator__(self, *, callback, **passkwargs):
        seed=f'{callback.__module__}.{callback.__name__}:{json.dumps(passkwargs, default=str, sort_keys=True)}'
        #seed=f'{json.dumps(passkwargs, default=str, sort_keys=True)}'
        assert self.pool == 1, 'use parallelindex for processing pool'
        for kwargs in self.kwargslist: 
            if not self.serialized(kwargs, seed): self.insert_hash(kwargs, seed, callback(**passkwargs, **kwargs))
            elif self.inmemory: self.update_hash(kwargs, seed, callback(**passkwargs, **kwargs))
            yield pickle.loads(self.serialized(kwargs, seed))

    def __exit__(self, exc_type, exc_value, tb):
        # TODO: 
        # if self.inmemory: copy in-memory database back to disk
        # https://stackoverflow.com/questions/4019081/how-to-copy-a-sqlite-table-from-a-disk-database-to-a-memory-database-in-python
        assert not self.inmemory, 'feature not yet implemented'  

    def insert_hash(self, kwargs={}, seed='', obj=None):
        logging.debug(f'INSERT HASH {self.hash_dict(kwargs, seed)}\nseed = {seed}\nBIN: kwargs = {kwargs}')
        with sqlite3.connect(self.storage) as con:
            db = con.cursor()
            db.execute('INSERT INTO hashmap VALUES (?,?)', (self.hash_dict(kwargs, seed), bytes(pickle.dumps(obj) if self.store else pickle.dumps(None))))

    def update_hash(self, kwargs={}, seed='', obj=None):
        logging.debug(f'UPDATE HASH {self.hash_dict(kwargs, seed)}\nseed = {seed }\nBIN: kwargs = {kwargs}')
        with sqlite3.connect(self.storage) as con:
            db = con.cursor()
            db.execute('UPDATE hashmap SET bytes = ? WHERE hash = ?', (pickle.dumps(obj), self.hash_dict(kwargs, seed)))

    def serialized(self, kwargs={}, seed=''):
        """ returns binary object or True if hash exists in database, else False """
        with sqlite3.connect(self.storage) as con:
            db = con.cursor()
            db.execute('SELECT * FROM hashmap WHERE hash == ?', (self.hash_dict(kwargs, seed),))
            res = db.fetchone()
        logging.debug(f'CHECK HASH {self.hash_dict(kwargs, seed)}: {"exists!" if res is not None else "missing!" }\nseed = {seed}\nBIN: kwargs = {kwargs}')
        if res is None: return False
        if res[1] is None: return True
        if res[1] is not None: return res[1]

    def bin_kwargs(self, dx, dy, dz, dt, **kwargs):
        """ generate argument sets as area subsets of boundary kwargs

            kwargs are split into dx° * dy° * dz° * dt bins.
            coordinates are rounded to nearest outer-boundary degree integer

            args:
                deltas: (float or timedelta)
                    bin size of axes 

                **kwargs (dict) 
                    boundary arguments, e.g.
                    >>> kwargs = {
                    ... 'south': 43.21, 'north': 46.54, 'west': -123.45, 'east': -110.01, 'top': 0, 'bottom': 5000, 
                    ... 'start': datetime(2000, 1, 1, 0, 0), 'end': datetime(2000, 1, 2, 0, 0)}

            yields: 
                dictionary containing args as a subset of input boundaries
        """
        if 'start' not in kwargs.keys(): kwargs['start'] = datetime(2000,1,1) 
        if 'end' not in kwargs.keys(): kwargs['end'] = kwargs['start']+dt

        for axmin, axmax, delta in zip(('west','south','bottom','start'), ('east','north','top','end'), (dx,dy,dz,dt)): 
            if axmin not in kwargs.keys(): kwargs[axmin] = 0
            if axmax not in kwargs.keys(): kwargs[axmax] = 0
            if min(kwargs[axmin],kwargs[axmax]) == max(kwargs[axmin],kwargs[axmax]): kwargs[axmax] += delta

        # spin to win!
        spacebins = lambda a,b,delta: arange(min(a,b)-(min(a,b)%(delta*1)), max(a,b)-(max(a,b)%(delta*-1)), delta)
        for x in spacebins(kwargs['west'], kwargs['east'], dx):
            for y in spacebins(kwargs['south'], kwargs['north'], dy): 
                for z in spacebins(kwargs['top'], kwargs['bottom'], dz):
                    for t in arange(kwargs['start'].date(), kwargs['end'], dt).astype(datetime):
                        yield dict(zip(('west', 'east', 'south', 'north', 'top', 'bottom', 'start', 'end',), (x, x+dx, y, y+dy, z, z+dz, t, t+dt,)))


class parallelindex(index):
    """ run index jobs in a parallel processing pool """

    def __call__(self, *, callback, **passkwargs):
        assert len(self.kwargslist) > 1, 'nothing to parallelize when bins=False'
        with Pool(self.pool) as p: 
            return list(p.map(self.__call_generator__, zip((callback for _ in self.kwargslist), self.kwargslist, (passkwargs for _ in self.kwargslist))))

    def __call_generator__(self, args):
        callback, kwargs, passkwargs = args
        seed=f'{callback.__module__}.{callback.__name__}:{json.dumps(passkwargs, default=str, sort_keys=True)}'
        if not self.serialized(kwargs, seed): self.insert_hash(kwargs, seed, callback(**passkwargs, **kwargs))
        elif self.inmemory: self.update_hash(kwargs, seed, callback(**passkwargs, **kwargs))
        return pickle.loads(self.serialized(kwargs, seed))


if __name__ == '__main__':
    """ run the demo to split boundary arguments into smaller areas, 
        then log the results of each function call

        >   python3.8 index.py 


        run demo again to quickly load cached results

        >   python3.8 index.py 
    """
    import time

    def callback(**kwargs):
        """ demo: some arbitrary slow process that accepts space/time boundaries as args """
        print(f'hello world!\n{json.dumps(kwargs, default=str, indent=1)}')
        time.sleep(0.5)
        return str(datetime.now().time())

    def parallelized_callback(**kwargs):
        """ another useless demo function """
        print(f'hello world! {kwargs}')
        time.sleep(1)
        return None

    # define some boundaries using or subsetting these dict keys (only used when bins=True)
    kwargs = {
            'west':    -123.45,     'east':    -110.01,
            'south':    43.21,      'north':    46.54, 
            'bottom':   5000,       'top':      0, 
            'start':    datetime(2000, 1, 1, 0, 0), 
            'end':      datetime(2000, 1, 2, 0, 0)
        }

    # here kwargs will split into 21 function calls using default spatial bin sizes
    with index(bins=True, store=True, inmemory=False, **kwargs) as scheduler: 
        results = scheduler(callback=callback, testarg='some arg', anotherarg='changing this will invalidate results hash')

    print(results)

    # and again, but this time in parallel
    with parallelindex(pool=10, **kwargs) as scheduler: 
        scheduler(callback=parallelized_callback, newargument='yerp')




import os
import logging
import zipfile
import requests
from functools import reduce
from datetime import datetime, timedelta

import numpy as np
from tqdm import tqdm

import kadlu
from kadlu import index
from kadlu.geospatial.data_sources.data_util import (
        database_cfg,
        logmsg,
        logmsg_nodata,
        storage_cfg,
    )


fname = 'GEBCO_2020.nc'

def initdb():
    conn, db = database_cfg()
    db.execute('CREATE TABLE IF NOT EXISTS gebco' 
                '(   val     REAL    NOT NULL,  ' 
                '    lat     REAL    NOT NULL,  ' 
                '    lon     REAL    NOT NULL  )' )
    db.execute('CREATE UNIQUE INDEX IF NOT EXISTS idx_gebco on gebco(val, lat, lon)')
    conn.close()


class Gebco():

    def fetch_bathymetry_grid(self):
        """ download geotiff zip archive and extract it """

        zipf = os.path.join(kadlu.storage_cfg(), "gebco_2020_netcdf.zip")

        # unzip the file if necessary
        if not os.path.isfile(os.path.join(storage_cfg(), fname)):

            # if there is no zip file, download it
            if not os.path.isfile(zipf):
                logging.info('downloading and decompressing gebco bathymetry (netcdf ~8GB)... ')
                url = 'https://www.bodc.ac.uk/data/open_download/gebco/gebco_2020/zip/'
                with requests.get(url, stream=True) as payload_netcdf:
                    assert payload_netcdf.status_code == 200, 'error fetching file'
                    with open(zipf, 'wb') as f:
                        with tqdm(total=4008688379, desc='gebco_2020_netcdf.zip', unit='B', unit_scale=True) as t:
                            for chunk in payload_netcdf.iter_content(chunk_size=8192): 
                                f.write(chunk)
                                _ = t.update(8192)

            # unzip the downloaded file
            with zipfile.ZipFile(zipf, 'r') as zip_ref:
                logging.info('extracting bathymetry data...')
                zip_ref.extractall(path=storage_cfg())
        return


    def fetch_callback(self, south, north, west, east, top=None, bottom=None, start=None, end=None):
        """ build data grid indexes from .nc grid and insert into the database """

        if not os.path.isfile(os.path.join(storage_cfg(), fname)):
            self.fetch_bathymetry_grid()

        rows = kadlu.load_file(os.path.join(storage_cfg(), fname), south=south, north=north, west=west, east=east).T

        initdb()
        conn, db = database_cfg()
        n1 = db.execute(f"SELECT COUNT(*) FROM gebco ").fetchall()[0][0]
        db.executemany(f"INSERT OR IGNORE INTO gebco (val, lat, lon)  VALUES (?,?,?)", rows)
        n2 = db.execute(f"SELECT COUNT(*) FROM gebco ").fetchall()[0][0]

        logmsg('gebco', 'bathymetry', (n1, n2), south=south, west=west, north=north, east=east)
        conn.commit()
        conn.close()
        return


    def load_bathymetry(self, south, north, west, east, start=datetime(2000,1,1), end=datetime(2000,1,2), top=0, bottom=99999, **_):
        """ load gebco bathymetry data """

        with index(dx=15, dy=5, dz=(bottom-top or 1), dt=(end-start or timedelta(1)), 
                    storagedir=storage_cfg(), bins=True, 
                    south=south, north=north, west=west, east=east,
                    top=top, bottom=bottom, start=start, end=end
                ) as fetchmap:
            _ = fetchmap(callback=self.fetch_callback)

        conn, db = kadlu.database_cfg()
        qry = ' AND '.join([f'SELECT val, lat, lon FROM gebco WHERE lat >= ?','lat <= ?','lon >= ?','lon <= ?'])
        #print(qry)
        logging.debug(f'query: {qry}')
        db.execute(qry, (south, north, west, east))
        res = np.array(db.fetchall()).T
        conn.close()
        if len(res) == 0: 
            logmsg_nodata('gebco', 'bathymetry', south=south, north=north, west=west, east=east)
            return np.array([[],[],[]])
        val, lat, lon = res
        return val, lat, lon




"""
    additional tools and utils used across data loading modules
"""

import configparser
import json
import logging
import os
import pickle
import sqlite3
import sys
import warnings
from contextlib import contextmanager, redirect_stdout, redirect_stderr
from datetime import datetime, timedelta
from functools import reduce
from io import StringIO
from os.path import dirname
from importlib import reload

import numpy as np


ext = lambda filepath, extensions: isinstance(extensions, tuple) and any(x.lower() == filepath.lower()[-len(x):] for x in extensions)


def database_cfg():
    """ configure and connect to sqlite database

        time is stored as an integer in the database, where each value
        is epoch hours since 2000-01-01 00:00

        returns:
            conn:   
                database connection object
            db:
                connection cursor object
    """
    conn = sqlite3.connect(storage_cfg() + 'geospatial.db')
    db = conn.cursor()

    return conn, db


class storage_cfg(os.PathLike):
    """ return filepath containing storage configuration string

        first checks the config.ini file in kadlu root folder, then
        defaults to kadlu/storage
    """
    """
    __file__  = '/home/matt_s/kadlu/kadlu/geospatial/data_sources/data_util.py'
    """
    cfg = configparser.ConfigParser()       # read .ini into dictionary object

    def __init__(self, setdir=None):
        self.cfg.read(os.path.join(dirname(dirname(dirname(dirname(__file__)))), "config.ini"))
        if setdir: self.__call__(setdir)

    def default_storage(self, msg):
        """ helper function for storage_cfg() """
        storage_location = os.path.join(os.path.expanduser('~'), f'kadlu_data{os.path.sep}')
        if not os.path.isdir(storage_location): os.mkdir(storage_location)
        warnings.warn(f'''{msg} storage location will be set to {storage_location}.\n'''
        '''use kadlu.storage_cfg('/path/to/new/directory') to configure a different storage location''')
        return storage_location

    def __call__(self, setdir=None) -> str:
        if 'storage' not in self.cfg.sections():
            self.cfg.add_section('storage')

        if setdir is not None:
            assert os.path.isdir(setdir), 'directory does not exist'
            self.cfg.set('storage', 'storage_location', setdir)
            with open(os.path.join(dirname(dirname(dirname(dirname(__file__)))), "config.ini"), 'w') as f:
                self.cfg.write(f)
        try:
            storage_location = self.cfg['storage']['storage_location']
        except KeyError:                        # missing config.ini file
            return self.default_storage('storage location not configured.')

        if storage_location == '':              # null value in config.ini
            return self.default_storage('null value in config.')

        if not os.path.isdir(storage_location):    # verify the location exists
            return self.default_storage('storage location does not exist.')

        if storage_location[-1] != os.path.sep: 
            return storage_location + os.path.sep

        return str(storage_location)

    def __repr__(self) -> str: 
        return self.__str__()

    def __str__(self) -> str: 
        return self.__call__()

    def __dir__(self) -> str: 
        return self.__call__()

    def __fspath__(self) -> str: 
        return self.__call__()

    def __add__(self, other) -> str: 
        return ''.join([self.__str__(), other.__str__()])


def verbosity(set_verbosity=None):
    '''
    __file__ = '/home/matt/kadlu/kadlu/geospatial/data_sources/data_util.py'
    '''
    cfg = configparser.ConfigParser()
    cfg.read(os.path.join(dirname(dirname(dirname(dirname(__file__)))), "config.ini"))

    if 'other' not in cfg.sections():
        cfg.add_section('other')
        cfg.set('other', 'verbosity', str(False))
        with open(os.path.join(dirname(dirname(dirname(dirname(__file__)))), "config.ini"), 'w') as f:
            cfg.write(f)
    
    if set_verbosity is not None:
        assert set_verbosity in (True, False), 'verbosity must be True or False'
        cfg.set('other', 'verbosity', str(set_verbosity))
        with open(os.path.join(dirname(dirname(dirname(dirname(__file__)))), "config.ini"), 'w') as f:
            cfg.write(f)

    return cfg['other']['verbosity'] == 'True'

def dt_2_epoch(dt_arr, t0=datetime(2000,1,1,0,0,0)):
    """ convert datetimes to epoch hours """
    delta = lambda dt: (dt - t0).total_seconds()/60/60
    if isinstance(dt_arr, (list, np.ndarray)): return list(map(int, map(delta, dt_arr)))
    elif isinstance(dt_arr, (datetime)): return int(delta(dt_arr))
    else: raise ValueError('input must be datetime or array of datetimes')


def epoch_2_dt(ep_arr, t0=datetime(2000,1,1,0,0,0), unit='hours'):
    """ convert epoch hours to datetimes """
    delta = lambda ep, unit: t0 + timedelta(**{f'{unit}' : ep})
    if isinstance(ep_arr, (list, np.ndarray)): return list(map(partial(delta, unit=unit), ep_arr))
    elif isinstance(ep_arr, (float, int)): return delta(ep_arr, unit=unit)
    else: raise ValueError('input must be integer or array of integers')


def index_arr(val, sorted_arr):
    """ converts value in coordinate array to grid index """
    if val > sorted_arr[-1]: return len(sorted_arr) -1
    return np.nonzero(sorted_arr >= val)[0][0]


def flatten(cols, frames):
    """ dimensional reduction by taking average of time frames """
    # assert that frames are of equal size
    assert len(cols) == 5
    assert reduce(lambda a, b: (a==b)*a, frames[1:] - frames[:-1])

    # aggregate time frame splits and reduce them to average
    frames_ix = np.append([0], frames)
    split_num = range(len(frames_ix) -1)
    val_split = np.array([cols[0][frames_ix[f] : frames_ix[f+1]] for f in split_num])
    value_avg = (reduce(np.add, val_split) / len(val_split))
    _, y, x, _, z = cols[:, frames_ix[0] : frames_ix[1]]
    return value_avg, y, x, z


def reshape_2D(cols):
    return dict(values=cols[0], lats=cols[1], lons=cols[2])


def reshape_3D(cols):
    """ prepare loaded data for interpolation 
    
        args:
            cols: flattened numpy array of shape (4, n)
                cols[0]: values
                cols[1]: latitude
                cols[2]: longitude
                cols[3]: depth

        return: gridded
            dict(values=gridspace, lats=ygrid, lons=xgrid, depths=zgrid)
    
    """
    if isinstance(cols[0], (float, int)):
        return dict(values=cols[0])

    # if data is 4D, take average of values at time frame intervals
    frames = np.append(np.nonzero(cols[3][1:] > cols[3][:-1])[0] + 1, len(cols[3]))
    if len(np.unique(frames)) > 1: vals, y, x, z = flatten(cols, frames) 
    else: vals, y, x, _, z  = cols
    rows = np.array((vals, y, x, z)).T

    # reshape row data to 3D array
    xgrid, ygrid, zgrid = np.unique(x), np.unique(y), np.unique(z)
    gridspace = np.full((len(ygrid), len(xgrid), len(zgrid)), fill_value=None, dtype=float)
    # this could potentially be optimized to avoid an index lookup cost
    for row in rows:
        x_ix = index_arr(row[2], xgrid)
        y_ix = index_arr(row[1], ygrid)
        z_ix = index_arr(row[3], zgrid)
        gridspace[y_ix, x_ix, z_ix] = row[0]

    # remove nulls for interpolation:
    # fill missing depth values with last value in each column
    for xi in range(0, gridspace.shape[0]):
        for yi in range(0, gridspace.shape[1]):
            col = gridspace[xi, yi]
            if sum(np.isnan(col)) > 0 and sum(np.isnan(col)) < len(col):
                col[np.isnan(col)] = col[~np.isnan(col)][-1]
                gridspace[xi, yi] = col

    # null depth columns are filled with the average value at each depth plane
    for zi in range(0, gridspace.shape[2]):
        gridspace[:,:,zi][np.isnan(gridspace[:,:,zi])] = np.average(gridspace[:,:,zi][~np.isnan(gridspace[:,:,zi])])

    return dict(values=gridspace, lats=ygrid, lons=xgrid, depths=zgrid)


def str_def(self, info, args):
    """ builds string definition for data source class objects """
    fcns = [fcn for fcn in dir(self) if callable(getattr(self, fcn)) 
            and 'load' in fcn and not fcn.startswith("__")]
    return (f'{info}\n\nfunction input arguments:\n\t{args}\n\nclass functions:\n\t'
            + '\n\t'.join(fcns) + '\n')


def ll_2_regionstr(south, north, west, east, regions, default=[]):
    """ convert input bounds to region strings with seperating axis theorem """
    if west > east:  # recursive function call if query intersects antimeridian
        return np.union1d(ll_2_regionstr(south, north, west, +180, regions, default), 
                          ll_2_regionstr(south, north, -180, east, regions, default))

    query = Boundary(south, north, west, east)
    matching = [str(reg) for reg in regions if query.intersects(reg)]

    if len(matching) == 0: 
        warnings.warn(f"No regions matched for query. Defaulting to {default} ({len(default)} regions)")
        return default

    return np.unique(matching)


def fmt_coords(kwargs):
      return (
            f"{abs(kwargs['south']):.2f}°{'S' if kwargs['south'] <= 0 else 'N'}," 
            f"{abs(kwargs['west']):.2f}°{'W' if kwargs['west'] <= 0 else 'E'}"
            f"{','+str(kwargs['top'])+'m' if 'top' in kwargs.keys() else ''} : "
            f"{abs(kwargs['north']):.2f}°{'S' if kwargs['north'] <= 0 else 'N'},"
            f"{abs(kwargs['east']):.2f}°{'W' if kwargs['east'] <= 0 else 'E'}"
            f"{','+str(kwargs['bottom'])+'m' if 'bottom' in kwargs.keys() else ''}"
        )


def logmsg(source, var, ntup=(), **kwargs):
    logging.info(f'{source.upper()} {var.lower()}  logged {reduce(np.subtract, ntup[::-1])} points in region\t{json.dumps(kwargs, default=str, sort_keys=True)}')


def logmsg_nodata(source, var, **kwargs):
    logging.warning(f'{source.upper()} {var.lower()}: no data found, returning empty arrays for region\t{json.dumps({k:v for k,v in kwargs.items() if k in ("north","south","west","east","top","bottom","start","end")} , default=str, sort_keys=True)}')



class Boundary():
    """ compute intersecting boundaries with separating axis theorem """
    def __init__(self, south, north, west, east, fetchvar='', **_):
        self.south, self.north, self.west, self.east, self.fetchvar = \
                south, north, west, east, fetchvar

    def __str__(self): return self.fetchvar

    def intersects(self, other):
        return not (self.east  < other.west or
                    self.west  > other.east or
                    self.north < other.south or
                    self.south > other.north)


class Capturing(list):
    # https://stackoverflow.com/questions/16571150/how-to-capture-stdout-output-from-a-python-function-call
    def __enter__(self):
        self._stdout = sys.stdout
        sys.stdout = self._stringio = StringIO()
        return self
    def __exit__(self, *args):
        self.extend(self._stringio.getvalue().splitlines())
        del self._stringio
        sys.stdout = self._stdout


@contextmanager
def dev_null():
    """ context manager to redirect output to /dev/null """
    with open(os.devnull, 'w') as null:
        try:
            with redirect_stderr(null) as err, redirect_stdout(null) as out: 
                yield (err, out)
        finally:
            sys.stdout = sys.__stdout__
            sys.stderr = sys.__stderr__




import os
from setuptools import setup, find_packages

setup(name='kadlu',
        version=os.environ.get('KADLUVERSION', '0.0.0'), 
        description="MERIDIAN Python package for ocean ambient noise modelling",
        url='https://gitlab.meridian.cs.dal.ca/public_projects/kadlu',
        author='Oliver Kirsebom, Matthew Smith',
        author_email='oliver.kirsebom@dal.ca, matthew.smith@dal.ca',
        license='GNU General Public License v3.0',
        packages=find_packages(exclude=('tests',)),
        install_requires=[
            'cdsapi',
            'cdsapi',
            'eccodes',
            'gsw',
            'imageio',
            'matplotlib',
            'mpl_scatter_density',
            'netcdf4',
            'numpy',
            'Pillow',
            'pygrib',   # DEPENDS ON eccodes binaries
            'pyproj',
            'pyqt5',
            'requests',
            'scipy',
            'tqdm',
            'xarray',
            ],
        setup_requires=['pytest-runner',],
        tests_require=['pytest','pytest-parallel'],
        include_package_data=True,
        exclude_package_data={'':['tests']},
        python_requires='>=3.6',
    )

